(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016)
Author:  Rene Thiemann <rene.thiemann@uibk.ac.at> (2016)
Author:  Akihisa Yamada <akihisa.yamada@uibk.ac.at> (2016)
License: LGPL (see file COPYING.LESSER)
*)
theory AC_DP_Framework
imports
  QDP_Framework
  "../AC_Rewriting/AC_Rewriting"
begin

type_synonym ('f, 'v) ac_dpp =
  "('f, 'v) trs \<times> ('f, 'v) trs \<times> ('f, 'v) trs \<times> ('f, 'v) trs \<times> ('f, 'v) trs"

fun min_iacchain :: "('f, 'v) ac_dpp \<Rightarrow> ('f, 'v) term seq \<Rightarrow> ('f, 'v) term seq \<Rightarrow> ('f, 'v) subst seq \<Rightarrow> bool" where
  "min_iacchain (P,Q,R,Rw,E) s t \<sigma> =
  ((\<forall>i. (s i, t i) \<in> P \<union> Q) \<and>
  (\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep (R \<union> Rw \<union> E))\<^sup>*) \<and>
  ((\<exists>\<^sub>\<infinity>i. (s i, t i) \<in> P) \<or>
   (\<exists>\<^sub>\<infinity>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> relto (rstep R) (rstep (R \<union> Rw \<union> E)))) \<and>
  (\<forall>i. SN_on (relstep (R \<union> Rw) E) {t i \<cdot> \<sigma> i}))"

declare min_iacchain.simps[simp del]

text \<open>The following characterization admits to reuse lemmas such as
  @{thm ichain_split} which are crucial for reduction pair processors, etc.\<close>
lemma min_iacchain_via_ichain: "min_iacchain (P,Q,R,Rw,E) s t \<sigma> \<longleftrightarrow>
  ichain (False, False, P, Q, {}, R, Rw \<union> E) s t \<sigma> \<and> (\<forall>i. SN_on (relstep (R \<union> Rw) E) {t i \<cdot> \<sigma> i})"
by (auto simp: min_iacchain.simps ichain.simps ac_simps)

lemma min_iacchain_split:
  assumes chain: "min_iacchain (P,Q,R,Rw,E) s t \<sigma>"
    and nchain: "\<not> min_iacchain (Ps \<inter> (P \<union> Q), P \<union> Q - Ps, Rs \<inter> (R \<union> Rw), R \<union> Rw - Rs, E) s t \<sigma>"
  shows "\<exists> i. min_iacchain (P - Ps, Q - Ps, R - Rs, Rw - Rs, E) (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  have Rw: "R \<union> Rw = (Rs \<inter> (R \<union> Rw)) \<union> (R \<union> Rw - Rs)" by auto
  have nchain: "\<not> ichain (False,False,Ps \<inter> (P \<union> Q), P \<union> Q - Ps, {}, Rs \<inter> (R \<union> Rw),  R \<union> Rw - Rs \<union> E) s t \<sigma>"
  proof
    assume ichain: "ichain (False,False,Ps \<inter> (P \<union> Q), P \<union> Q - Ps, {}, Rs \<inter> (R \<union> Rw),  R \<union> Rw - Rs \<union> E) s t \<sigma>"
    hence "min_iacchain (Ps \<inter> (P \<union> Q), P \<union> Q - Ps, Rs \<inter> (R \<union> Rw), R \<union> Rw - Rs, E) s t \<sigma>"
      unfolding min_iacchain_via_ichain
      using chain[unfolded min_iacchain_via_ichain]
      by (simp add: Rw[symmetric])
    with nchain show False by auto
  qed
  from chain have ichain: "ichain (False,False,P,Q,{},R,Rw \<union> E) s t \<sigma>" 
    unfolding min_iacchain_via_ichain by auto
  from ichain_split_gen[OF ichain nchain] obtain i
    where ichain: "ichain (False,False,P - Ps, Q - Ps, {}, R - Rs, Rw - Rs \<union> E) (shift s i) (shift t i) (shift \<sigma> i)" ..
  have "min_iacchain (P - Ps, Q - Ps, R - Rs, Rw - Rs, E) (shift s i) (shift t i) (shift \<sigma> i)"
    unfolding min_iacchain_via_ichain
    by (rule conjI[OF ichain], rule allI,
    rule SN_on_mono[OF _ relto_mono[OF rstep_mono[of _ "R \<union> Rw"] subset_refl]],
    insert chain[unfolded min_iacchain_via_ichain], auto)
  thus ?thesis ..
qed

lemma min_iacchain_split_top:
  assumes chain: "min_iacchain (P,Q,R,Rw,E) s t \<sigma>"
    and nchain: "\<not> min_iacchain (Ps \<inter> (P \<union> Q), P \<union> Q - Ps, {}, R \<union> Rw, E) s t \<sigma>"
  shows "\<exists> i. min_iacchain (P - Ps, Q - Ps, R, Rw, E) (shift s i) (shift t i) (shift \<sigma> i)"
using min_iacchain_split [OF chain, of Ps "{}"] nchain by auto

definition "min_iacchain_alt P Q R Rw E s t \<longleftrightarrow>
  (\<forall>i. (s i, t i) \<in> rrstep (P \<union> Q)) \<and>
  (\<forall>i. (t i, s (Suc i)) \<in> (rstep (R \<union> Rw \<union> E))\<^sup>*) \<and>
  ((\<exists>\<^sub>\<infinity>i. (s i, t i) \<in> rrstep P) \<or>
   (\<exists>\<^sub>\<infinity>i. (t i, s (Suc i)) \<in> relstep R (R \<union> Rw \<union> E))) \<and>
  (\<forall>i. SN_on (relstep (R \<union> Rw) E) {t i})"

lemma min_iacchain_altI:
  "min_iacchain (P,Q,R,Rw,E) s t \<sigma> \<Longrightarrow> min_iacchain_alt P Q R Rw E (\<lambda>i. s i \<cdot> \<sigma> i) (\<lambda>i. t i \<cdot> \<sigma> i)"
by (auto simp: min_iacchain.simps min_iacchain_alt_def INFM_mono rrstepI)

lemma min_iacchain_altD:
  assumes "min_iacchain_alt P Q R Rw E s t"
  shows "\<exists>s t \<sigma>. min_iacchain (P,Q,R,Rw,E) s t \<sigma>"
proof -
  { fix i :: nat
    have "\<exists>u v \<sigma>. (u, v) \<in> P \<union> Q \<and> s i = u \<cdot> \<sigma> \<and> t i = v \<cdot> \<sigma>"
      using assms by (auto simp: min_iacchain_alt_def elim!: rrstepE dest!: spec [of _ i]) }
  then obtain u and v and \<sigma>
    where *: "\<forall>i. (u i, v i) \<in> P \<union> Q \<and> s i = u i \<cdot> \<sigma> i \<and> t i = v i \<cdot> \<sigma> i" by metis
  consider (P) "\<exists>\<^sub>\<infinity>i. (s i, t i) \<in> rrstep P" | (R) "\<exists>\<^sub>\<infinity>i. (t i, s (Suc i)) \<in> relstep R (R \<union> Rw \<union> E)"
    using assms by (auto simp: min_iacchain_alt_def)
  then show ?thesis
  proof (cases)
    def I \<equiv> "\<lambda>i. LEAST j. j \<ge> i \<and> (\<exists>u v \<sigma>. (u, v) \<in> P \<and> s j = u \<cdot> \<sigma> \<and> t j = v \<cdot> \<sigma>)" (is "\<lambda>i. LEAST j. ?P i j")
    case P
    then have P: "\<forall>i. \<exists>j. ?P i j" by (force simp: INFM_nat_le elim!: rrstepE)
    then have "\<forall>i. \<exists>u v \<sigma>. (u, v) \<in> P \<and> s (I i) = u \<cdot> \<sigma> \<and> t (I i) = v \<cdot> \<sigma>"
      by (auto simp: I_def dest!: LeastI_ex)
    then obtain u' and v' and \<sigma>'
      where **: "\<forall>i. (u' i, v' i) \<in> P \<and> s (I i) = u' i \<cdot> \<sigma>' i \<and> t (I i) = v' i \<cdot> \<sigma>' i" by metis
    def U \<equiv> "\<lambda>i. if \<exists>j. i = I j then u' (LEAST j. i = I j) else u i"
    def V \<equiv> "\<lambda>i. if \<exists>j. i = I j then v' (LEAST j. i = I j) else v i"
    def S \<equiv> "\<lambda>i. if \<exists>j. i = I j then \<sigma>' (LEAST j. i = I j) else \<sigma> i"
    { fix i :: nat
      have "(U i, V i) \<in> P \<union> Q \<and> (V i \<cdot> S i, U (Suc i) \<cdot> S (Suc i)) \<in> (rstep (R \<union> Rw \<union> E))\<^sup>* \<and>
        SN_on (relstep (R \<union> Rw) E) {V i \<cdot> S i}"
        using assms and * and **
        by (auto simp: min_iacchain_alt_def U_def V_def S_def) (metis (mono_tags, lifting) LeastI_ex)+ }
    moreover
    { fix i :: nat
      have "(U (I i), V (I i)) \<in> P" using ** by (auto simp: U_def V_def)
      moreover have "I i \<ge> i" using P by (auto simp: I_def dest!: LeastI_ex)
      ultimately have "\<exists>j\<ge>i. (U j, V j) \<in> P" by blast }
    ultimately show ?thesis by (auto simp: min_iacchain.simps INFM_nat_le)
  next
    case R
    then have "min_iacchain (P,Q,R,Rw,E) u v \<sigma>"
      using assms and * by (auto simp: min_iacchain_alt_def min_iacchain.simps)
    then show ?thesis by blast
  qed
qed

lemma no_chain_imp_SN_rel_ext:
  assumes " \<not> (\<exists>s t \<sigma>. min_iacchain (P, Q, R, Rw, E) s t \<sigma>)"
  shows "SN_rel_ext (rrstep P) (rrstep Q) (rstep R) (rstep Rw \<union> rstep E) (\<lambda>t. SN_on (relstep (R \<union> Rw) E) {t})"
    (is "SN_rel_ext ?P ?Q ?R (?Rw \<union> ?E) ?M")
proof (rule ccontr)
  let ?rel = "SN_rel_ext_step ?P ?Q ?R (?Rw \<union> ?E)"
  assume "\<not> ?thesis"
  from this[unfolded SN_rel_ext_def] obtain f ty where
    steps: "\<And> i. (f i, f (Suc i)) \<in> ?rel (ty i)"
   and min: "\<And> i. ?M (f i)"
   and inf1: "INFM i. ty i \<in> {top_s,top_ns}"
   and inf2: "INFM i. ty i \<in> {top_s,normal_s}" by blast
  obtain p where p: "\<And> i. p i = (ty i \<in> {top_s,top_ns})" by auto
  interpret infinitely_many p
    by (unfold_locales, insert inf1, unfold INFM_nat_le p)
  let ?ind = "infinitely_many.index p"
  let ?g = "\<lambda> i. f (?ind i)"
  let ?h = "\<lambda> i. f (Suc (?ind i))"
  let ?prop = "\<lambda> s t \<sigma> i. (s, t) \<in> P \<union> Q \<and> ?g i = s \<cdot> \<sigma> \<and> ?h i = t \<cdot> \<sigma> \<and> (ty (?ind i) = top_s \<longrightarrow> (s, t) \<in> P)"
  {
    fix i
    have "ty (?ind i) \<in> {top_s,top_ns}"
      using index_p[of i] unfolding p .
    hence one: "(?g i, ?h i) \<in> ?P \<union> ?Q" and two: "ty (?ind i) = top_s \<Longrightarrow> (?g i, ?h i) \<in> ?P" using steps[of "?ind i"]
      by auto      
    have "\<exists> s t \<sigma>. ?prop s t \<sigma> i"
    proof (cases "ty (?ind i) = top_s")
      case True
      with two[OF True] have "(?g i, ?h i) \<in> rrstep P" by simp
      then show ?thesis by (auto elim!: rrstepE)
    next
      case False
      moreover from one have "(?g i, ?h i) \<in> rrstep (P \<union> Q)" by (auto simp: rrstep_union)
      ultimately show ?thesis by (auto elim!: rrstepE)
    qed
  }
  from choice[OF allI[OF this]] obtain s where "\<forall> i. \<exists> t \<sigma>. ?prop (s i) t \<sigma> i" ..
  from choice[OF this] obtain t where "\<forall> i. \<exists> \<sigma>. ?prop (s i) (t i) \<sigma> i" ..
  from choice[OF this] obtain \<sigma> where st\<sigma>: "\<And> i. ?prop (s i) (t i) (\<sigma> i) i" by auto
  from st\<sigma> have PQ: "\<And> i. (s i, t i) \<in> P \<union> Q" by auto
  from st\<sigma> have s: "\<And> i. s i \<cdot> \<sigma> i = ?g i" by auto
  from st\<sigma> have t: "\<And> i. t i \<cdot> \<sigma> i = ?h i" by auto
  from st\<sigma> have P: "\<And> i. ty (?ind i) = top_s \<Longrightarrow> (s i, t i) \<in> P" by auto
  {
    fix i
    from st\<sigma>[of i] have "?M (t i \<cdot> \<sigma> i)" using min[of "Suc (?ind i)"] by auto
  } note min = this
  let ?f = "\<lambda> i j. f (Suc (?ind i) + j)"
  let ?n = "\<lambda> i. ?ind (Suc i) - Suc (?ind i)"
  let ?RE = "qrstep False {} (R \<union> (Rw \<union> E))"
  have "ichain (False, False, P, Q, {}, R, Rw \<union> E) s t \<sigma>"
    unfolding ichain_alternative
  proof (rule exI[of _ ?f], rule exI[of _ ?n], intro conjI allI impI, rule PQ)
    fix i
    show "?f i 0 = t i \<cdot> \<sigma> i" unfolding t by simp
  next
    fix i
    show "?f i (?n i) = s (Suc i) \<cdot> \<sigma> (Suc i)" unfolding s
      using index_ordered[of i] by auto
  next
    fix i j
    let ?k = "Suc (?ind i) + j"
    assume "j < ?n i"
    with index_not_p_between[of i ?k]
    have "\<not> p ?k" by auto
    hence "ty ?k \<in> {normal_s,normal_ns}" unfolding p by (cases "ty ?k", auto)
    with steps[of ?k] show "(?f i j, ?f i (Suc j)) \<in> ?RE" unfolding qrstep_union
      by auto
  next
    let ?L = "\<lambda> i. (s i, t i) \<in> P"
    let ?R = "\<lambda> i. \<exists> j < ?n i. (?f i j, ?f i (Suc j)) \<in> qrstep False {} R"
    show "(INFM i. ?L i) \<or> (INFM i. ?R i)" 
      unfolding INFM_disj_distrib[symmetric] 
      unfolding INFM_nat_le
    proof (intro allI)
      fix i
      from inf2[unfolded INFM_nat_le]
      obtain k where k: "k \<ge> ?ind i" and ty: "ty k \<in> {top_s,normal_s}" by auto
      from index_surj[OF k]
      obtain j j' where kj: "k = ?ind j + j'" and j': "?ind j + j' < ?ind (Suc j)" by auto
      note ty = ty[unfolded kj]
      from k[unfolded kj] j' have lt: "?ind i < ?ind (Suc j)" by auto
      {
        assume "j < i"
        hence "Suc j \<le> i" by auto
        from index_ordered_le[OF this] lt have False by auto
      }
      hence j: "i \<le> j" by presburger
      show "\<exists> j \<ge> i. ?L j \<or> ?R j"
      proof (intro exI conjI, rule j)
        show "?L j \<or> ?R j"
        proof (cases j')
          case 0
          from index_p[of j, unfolded p]
            ty[unfolded 0] have "ty (?ind j) = top_s" by auto
          from P[OF this] show ?thesis by auto
        next
          case (Suc j'')
          from index_not_p_between[of j "?ind j + j'", OF _ j']
          have "ty (?ind j + j') \<notin> {top_s,top_ns}" unfolding p Suc by auto
          with ty have ty: "ty (?ind j + j') = normal_s" by auto
          have j'': "j'' < ?ind (Suc j) - Suc (?ind j)" using j' unfolding Suc by auto
          have "?R j"
            by (rule exI[of _ j''], rule conjI[OF j''],
              insert steps[of "?ind j + j'", unfolded ty], unfold Suc, auto)
          thus ?thesis by auto
        qed
      qed
    qed
  qed auto
  then have "min_iacchain (P,Q,R,Rw,E) s t \<sigma>"
    using min unfolding min_iacchain_via_ichain by auto
  with assms show False by auto
qed

definition finite_ac_dpp :: "('f,'v)ac_dpp \<Rightarrow> bool" where "finite_ac_dpp DPP = (\<not>(\<exists>s t \<sigma>. min_iacchain DPP s t \<sigma>))"

lemma finite_dpp_via_finite_ac_dpp: "finite_dpp (nfs,True,P,Pw,{},R,Rw) = finite_ac_dpp (P,Pw,R,Rw,{})"
  unfolding finite_dpp_def finite_ac_dpp_def min_iacchain_via_ichain[abs_def] min_ichain.simps[abs_def]
  by (auto simp: minimal_cond_def ichain.simps)

lemma finite_ac_dppI[intro]: "(\<And> s t \<sigma>.  min_iacchain dpp s t \<sigma> \<Longrightarrow> False) \<Longrightarrow> finite_ac_dpp dpp"
  unfolding finite_ac_dpp_def by auto

lemma finite_ac_dpp_split_top: "finite_ac_dpp (P - Ps, Q - Ps, R, Rw, E) \<Longrightarrow> 
   finite_ac_dpp (Ps \<inter> (P \<union> Q), P \<union> Q - Ps, {}, R \<union> Rw, E) \<Longrightarrow>
   finite_ac_dpp (P,Q,R,Rw,E)"
   unfolding finite_ac_dpp_def using min_iacchain_split_top by blast

lemma finite_ac_dpp_split: "finite_ac_dpp (P - Ps, Q - Ps, R - Rs, Rw - Rs, E) \<Longrightarrow> 
  finite_ac_dpp (Ps \<inter> (P \<union> Q), P \<union> Q - Ps, Rs \<inter> (R \<union> Rw), R \<union> Rw - Rs, E) \<Longrightarrow>
  finite_ac_dpp (P,Q,R,Rw,E)"
  unfolding finite_ac_dpp_def using min_iacchain_split by blast

lemma finite_ac_dpp_pairs_mono: assumes fin: "finite_ac_dpp (P',Q',R,Rw,E)"
  and P: "P \<subseteq> P'" and Q: "Q \<subseteq> P' \<union> Q'"
  shows "finite_ac_dpp (P,Q,R,Rw,E)"
proof -
  from P Q have Q: "P \<union> Q \<subseteq> P' \<union> Q'" by auto
  have *: "\<And> x. x \<subseteq> x" by auto
  from fin show ?thesis
  unfolding finite_ac_dpp_def min_iacchain_via_ichain[abs_def] 
  using ichain_mono[OF _ P Q * * *, of False False "{}" R "Rw \<union> E"] by blast
qed


end