section{* Fresh variable addition *}

theory Fresh_Variable_Addition
  imports Cooperation_Program
begin

context pretheory begin

interpretation indexed_rewriting transition_step.

lemma assignment_update[intro]:
  "assignment \<alpha> \<Longrightarrow> a \<in> type_values ty \<Longrightarrow> assignment (\<alpha>((x,ty):=a))"
  using assms unfolding assignment_def by auto

definition update_state ("_ |\<^sub>s _ := _")
  where "update_state s x a = State ((valuation s)(x:=a)) (location s)"

lemma location_update[simp]: "location (s |\<^sub>s x := a) = location s"
  and valuation_update[simp]: "valuation (s |\<^sub>s x := a) = (valuation s)(x:=a)"
  by (auto simp: update_state_def)

lemma state_update:
  "state s \<Longrightarrow> a \<in> type_values ty \<Longrightarrow> state (s |\<^sub>s (x,ty):=a)"
  unfolding state_def valuation_update by(rule assignment_update)

definition change_fresh :: "'v \<Rightarrow> 't \<Rightarrow> (('v,'t,'d,'l) state \<times> ('v,'t,'d,'l) state) set"
where "change_fresh x ty \<equiv> { (s, s |\<^sub>s (x,ty) := a ) | s a. state s \<and> a \<in> type_values ty }"

lemma mem_change_freshI[intro]:
  assumes a: "a \<in> type_values ty"
      and s: "state s"
      and l: "location s = location t"
      and v: "valuation t = (valuation s)((x,ty):=a)"
  shows "(s,t) \<in> change_fresh x ty"
  unfolding change_fresh_def mem_Collect_eq prod.inject
proof(intro exI conjI)
  show "t = (s |\<^sub>s (x, ty) := a)" unfolding update_state_def l v[symmetric] by auto
qed (insert a s, auto)

lemma mem_change_freshD[dest]:
  assumes "(s,t) \<in> change_fresh x ty"
  shows "location s = location t"
    and "state s"
    and "state t"
    and "\<exists>a \<in> type_values ty. valuation t = (valuation s)((x,ty):=a)"
  using assms[unfolded change_fresh_def] state_update by auto

lemma change_fresh_sym:
  assumes "(s,t) \<in> change_fresh x ty"
  shows "(t,s) \<in> change_fresh x ty"
proof
  from assms obtain a
  where a: "a \<in> type_values ty" and ts: "valuation t = (valuation s)((x,ty):=a)" by auto
  with assms
  show "state t"
   and "valuation s = (valuation t)((x,ty):= valuation s (x,ty))"
   and "location t = location s" by auto
  from assms have "state s" by auto
  then show "valuation s (x, ty) \<in> local.type_values ty" by auto
qed

lemma change_fresh_refl_on:
  shows "refl_on (Collect state :: ('v, 't, 'd, 'l) state set) (change_fresh x ty)"
proof(rule refl_onI)
  fix s :: "('v, 't, 'd, 'l) state" assume s: "s \<in> Collect state"
  show "(s,s) \<in> change_fresh x ty"
  proof
    from s show "state s" by auto
    then show "valuation s (x,ty) \<in> type_values ty" by auto
  qed auto
qed auto

lemma change_fresh_trans: "trans (change_fresh x ty)" (is "trans ?S")
proof(rule transI)
  fix s t u assume st: "(s,t) \<in> ?S" and tu: "(t,u) \<in> ?S"
  show "(s,u) \<in> ?S"
  proof
    from tu show "valuation u (x,ty) \<in> type_values ty" by auto
  qed(insert mem_change_freshD[OF st] mem_change_freshD[OF tu], auto)
qed



definition "fresh_variable_cond P P' x ty \<equiv>
  lts.initial P' = lts.initial P \<and>
  (\<forall> \<tau> \<in> transitions P. \<exists> \<tau>' \<in> transitions P'.
   source \<tau>' = source \<tau> \<and> target \<tau>' = target \<tau> \<and>
   (\<forall>\<alpha>. assignment \<alpha> \<longrightarrow> \<alpha> \<Turnstile> transition.formula \<tau> \<longrightarrow>
     (\<exists>a \<in> type_values ty. \<alpha>((Post x,ty) := a) \<Turnstile> transition.formula \<tau>'))) \<and>
  (\<forall> \<tau>' \<in> transitions P'. \<forall> (s,t) \<in> transition_step \<tau>'. \<forall> a \<in> type_values ty.
   \<exists> b \<in> type_values ty. (s |\<^sub>s (x,ty):=a, t |\<^sub>s (x,ty):=b) \<in> transition_step \<tau>')"

context fixes P P' x ty assumes fv: "fresh_variable_cond P P' x ty" begin
qualified lemma D1: "lts.initial P' = lts.initial P"
  using fv[unfolded fresh_variable_cond_def] by auto

qualified lemma D2:
  assumes "\<tau> \<in> transitions P"
  shows "\<exists> \<tau>' \<in> transitions P'.
    source \<tau>' = source \<tau> \<and> target \<tau>' = target \<tau> \<and>
    (\<forall>\<alpha>. assignment \<alpha> \<longrightarrow> \<alpha> \<Turnstile> transition.formula \<tau> \<longrightarrow>
       (\<exists>a \<in> type_values ty. \<alpha>((Post x,ty) := a) \<Turnstile> transition.formula \<tau>'))"
  using assms fv[unfolded fresh_variable_cond_def] by auto

qualified lemma D3:
  "\<tau>' \<in> transitions P' \<Longrightarrow> (s,t) \<in> transition_step \<tau>' \<Longrightarrow> a \<in> type_values ty \<Longrightarrow>
   \<exists> b \<in> type_values ty. (s |\<^sub>s (x,ty):=a, t |\<^sub>s (x,ty):=b) \<in> transition_step \<tau>'"
  using fv[unfolded fresh_variable_cond_def, THEN conjunct2, THEN conjunct2, rule_format] by auto

lemmas fresh_variable_condD = D1 D2 D3

lemma fresh_variable_cond_imp_step:
  assumes "\<tau> \<in> transitions P"
  shows "\<exists>\<tau>' \<in> transitions P'.
    source \<tau>' = source \<tau> \<and>
    target \<tau>' = target \<tau> \<and>
    (\<forall>(s,t) \<in> transition_step \<tau>. \<exists>a \<in> type_values ty. (s, t |\<^sub>s (x,ty):=a) \<in> transition_step \<tau>')"
proof-
  let ?ppi = "\<lambda>s t. pre_post_inter  (valuation s) (valuation t)"
  from D2[OF assms(1)] obtain \<tau>'
  where \<tau>': "\<tau>' \<in> transitions P'"
    and [simp]: "source \<tau>' = source \<tau>" "target \<tau>' = target \<tau>"
    and 1: "\<And>\<alpha>. assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> transition.formula \<tau> \<Longrightarrow>
        (\<exists>a \<in> type_values ty. \<alpha>((Post x, ty) := a) \<Turnstile> transition.formula \<tau>')" by auto
  show ?thesis
  proof (intro bexI[OF _ \<tau>'] conjI ballI2)
    fix s t assume "(s,t) \<in> transition_step \<tau>"
    then obtain \<alpha>
    where \<alpha>: "assignment \<alpha>"
      and s: "state s"
      and t: "state t"
      and 2: "?ppi s t \<alpha> \<Turnstile> transition.formula \<tau>" (is "?a \<Turnstile> _")
      and [simp]: "source \<tau> = location s" "target \<tau> = location t"
      by auto
    from s t \<alpha> have "assignment ?a" by auto
    from 1[OF this 2] obtain a
    where a: "a \<in> type_values ty" and *: "?a((Post x, ty) := a) \<Turnstile> transition.formula \<tau>'" by auto
    note *
    also have "(?a((Post x, ty) := a)) xx = ?ppi s (t |\<^sub>s (x,ty):=a) \<alpha> xx" for xx by(cases xx, cases "fst xx",auto)
    then have "?a((Post x, ty) := a) = ?ppi s (t |\<^sub>s (x,ty):=a) \<alpha>" by auto
    finally have "... \<Turnstile> transition.formula \<tau>'" by auto
    with s t \<alpha> a show "\<exists>a\<in>local.type_values ty. (s, t |\<^sub>s (x, ty) := a) \<in> transition_step \<tau>'"
      by (intro bexI[OF _ a] mem_transition_stepI, auto)
  qed auto
qed


end

context
  fixes P P' :: "('f,'v,'t,'l sharp) lts" and x :: 'v and ty :: 't
  assumes fv: "fresh_variable_cond P P' x ty"
begin

lemma change_fresh_O_transition_step:
  assumes \<tau>': "\<tau>' \<in> transitions P'"
  shows "change_fresh x ty O transition_step \<tau>' \<subseteq> transition_step \<tau>' O change_fresh x ty" (is "?L \<subseteq> ?R")
proof
  fix s u assume L: "(s,u) \<in> ?L"
  then obtain t where st: "(s,t) \<in> change_fresh x ty" and tu: "(t,u) \<in> transition_step \<tau>'" by auto
  let ?a = "valuation s (x,ty)"
  from st have s: "assignment (valuation s)" and t: "assignment (valuation t)" by auto
  then have a': "?a \<in> type_values ty" by auto
  from tu have u: "assignment (valuation u)" by auto
  from st obtain a where "valuation t = (valuation s)((x,ty):=a)" by auto
  moreover have "...((x,ty):= valuation s (x,ty)) = valuation s" by auto
  moreover from st have "location t = location s" by auto
  ultimately have "(t |\<^sub>s (x,ty) := ?a) = s" unfolding update_state_def by (cases t, auto)
  from fresh_variable_condD(3)[OF fv \<tau>' tu a', unfolded this]
  obtain b where b: "b \<in> type_values ty" and sv: "(s, u |\<^sub>s (x, ty) := b) \<in> transition_step \<tau>'" by auto
  from b u have vu: "(u |\<^sub>s (x, ty) := b, u) \<in> change_fresh x ty" by (subst change_fresh_sym, auto)
  from sv vu show "(s,u) \<in> ?R" by auto
qed

lemma change_fresh_push:
  assumes \<tau>': "\<tau>' \<in> transitions P'"
  shows "change_fresh x ty O transition_step \<tau>' \<subseteq> transition_step \<tau>' O (change_fresh x ty)\<^sup>*" (is "?L \<subseteq> ?R O ?S\<^sup>*")
proof-
  from change_fresh_O_transition_step[OF assms]
  have "?L \<subseteq> ?R O ?S".
  also have "... \<subseteq> ?R O ?S\<^sup>*" by auto
  finally show ?thesis.
qed

lemma change_fresh_O_traverse:
  assumes \<tau>s': "\<tau>s' \<subseteq> transitions P'"
  shows "change_fresh x ty O traverse \<tau>s' \<subseteq> traverse \<tau>s' O (change_fresh x ty)\<^sup>*" (is "?S O ?R \<subseteq> _")
proof-
  have "?S O ?R \<subseteq> ?S\<^sup>* O ?R" by auto
  also have "... \<subseteq> ?R O ?S\<^sup>*" using \<tau>s' by (intro traverse_push change_fresh_push, auto)
  finally show ?thesis.
qed

lemma change_fresh_O_flat_transitions:
  shows "change_fresh x ty O transitions_step (flat_transitions_of P') \<subseteq> transitions_step (flat_transitions_of P') O (change_fresh x ty)\<^sup>*"
    (is "?L \<subseteq> ?R")
proof
  fix s t assume "(s,t) \<in> ?L"
  then obtain \<tau>
  where *: "\<tau> \<in> flat_transitions_of P'" "(s,t) \<in> change_fresh x ty O transition_step \<tau>" by auto
  then have "\<tau> \<in> transitions P'" by auto
  from * change_fresh_O_transition_step[OF this] have "(s,t) \<in> transition_step \<tau> O change_fresh x ty" by auto
  with * show "(s,t) \<in> ?R" by auto
qed


lemma transition_simulate:
  assumes \<tau>: "\<tau> \<in> transitions P"
  shows "\<exists> \<tau>' \<in> transitions P'.
    source \<tau>' = source \<tau> \<and>
    target \<tau>' = target \<tau> \<and>
    transition_step \<tau> \<subseteq> transition_step \<tau>' O (change_fresh x ty)\<^sup>*"
proof-
  from fresh_variable_cond_imp_step[OF fv \<tau>]
  obtain \<tau>'
  where \<tau>': "\<tau>' \<in> transitions P'"
    and [simp]: "source \<tau>' = source \<tau>" "target \<tau>' = target \<tau>"
    and *: "\<forall>(s,t) \<in> transition_step \<tau>. \<exists>a\<in>local.type_values ty.
      (s, t |\<^sub>s (x, ty) := a) \<in> transition_step \<tau>'" by auto
  { fix s t assume st: "(s,t) \<in> transition_step \<tau>"
    from st have vs: "assignment (valuation s)" and vt: "assignment (valuation t)" by auto
    then have a': "valuation t (x,ty) \<in> type_values ty" by auto
    from * st obtain a
    where a: "a \<in> type_values ty" and st': "(s, t |\<^sub>s (x,ty) := a) \<in> transition_step \<tau>'" by auto
    moreover with vt have "(t |\<^sub>s (x,ty) := a, t) \<in> change_fresh x ty"
      by (intro mem_change_freshI[OF a'] state_update, auto)
    ultimately have "(s,t) \<in> transition_step \<tau>' O change_fresh x ty" by auto
    also have "... \<subseteq> transition_step \<tau>' O (change_fresh x ty)\<^sup>*" by auto
    finally have "(s,t) \<in> ...".
  }
  then show ?thesis by(intro bexI[OF _ \<tau>'],auto)
qed

lemma flat_transitions_simulate:
  shows "transitions_step (flat_transitions_of P) \<subseteq> transitions_step (flat_transitions_of P') O (change_fresh x ty)\<^sup>*"
    (is "?L \<subseteq> ?R")
proof
  fix s t assume "(s,t) \<in> ?L"
  then obtain \<tau> where *: "\<tau> \<in> flat_transitions_of P" "(s,t) \<in> transition_step \<tau>" by auto
  with transition_simulate obtain \<tau>'
  where 1: "\<tau>' \<in> transitions P'"
    and 2: "transition_step \<tau> \<subseteq> transition_step \<tau>' O (change_fresh x ty)\<^sup>*" by auto
  from * 2 have 3: "(s,t) \<in> transition_step \<tau>' O (change_fresh x ty)\<^sup>*" by auto
  then have "source \<tau>' = location s" by auto
  moreover from * have "is_flat (location s)" by auto
  ultimately have "is_flat_transition \<tau>'" by auto
  with 1 have "\<tau>' \<in> flat_transitions_of P'" by auto
  with 3 show "(s,t) \<in> ?R" by auto
qed

lemma sharp_transition_simulate:
  assumes \<tau>: "\<tau> \<in> sharp_transitions_of P"
  shows "\<exists> \<tau>' \<in> sharp_transitions_of P'. transition_step \<tau> \<subseteq> transition_step \<tau>' O (change_fresh x ty)\<^sup>*"
    (is "\<exists> \<tau>' \<in> _. ?goal \<tau>'")
proof-
  from \<tau> have "\<tau> \<in> transitions P" by auto
  from transition_simulate[OF this] obtain \<tau>'
  where 1: "\<tau>' \<in> transitions P'"
    and 2: "source \<tau>' = source \<tau>"
    and 3: "?goal \<tau>'" by auto
  from \<tau> 1 2 have "\<tau>' \<in> sharp_transitions_of P'" by auto
  with 3 show ?thesis by auto
qed

lemma sharp_transition_simulate_fun:
  "\<exists>f. \<forall>\<tau> \<in> sharp_transitions_of P.
   f \<tau> \<in> sharp_transitions_of P' \<and> transition_step \<tau> \<subseteq> transition_step (f \<tau>) O (change_fresh x ty)\<^sup>*"
  using sharp_transition_simulate by (intro bchoice, auto)

lemma sharp_traverse_simulate:
  assumes \<tau>s: "\<tau>s \<subseteq> sharp_transitions_of P" and nonemp: "\<tau>s \<noteq> {}"
  shows "\<exists>\<tau>s' \<subseteq> sharp_transitions_of P'. \<tau>s' \<noteq> {} \<and> traverse \<tau>s \<subseteq> traverse \<tau>s' O (change_fresh x ty)\<^sup>*"
proof-
  let ?S = "change_fresh x ty"
  interpret I': indexed_rewriting "(\<lambda>\<tau>'. transition_step \<tau>' O ?S\<^sup>*)".

  from sharp_transition_simulate_fun obtain f
  where dom: "\<forall>\<tau> \<in> sharp_transitions_of P. f \<tau> \<in> sharp_transitions_of P'"
  and mono: "\<forall>\<tau> \<in> sharp_transitions_of P. transition_step \<tau> \<subseteq> transition_step (f \<tau>) O ?S\<^sup>*" by auto
  with transition_simulate
  interpret rule_morphism id "sharp_transitions_of P" f transition_step "(\<lambda>\<tau>'. transition_step \<tau>' O ?S\<^sup>*)"
    by (unfold_locales,auto)

  note recurring_morphism

  interpret I: indexed_rewriting "(\<lambda>\<tau>. transition_step (f \<tau>) O ?S\<^sup>*)".


  have *: "I.traversed \<tau>s seq i \<Longrightarrow> I'.traversed (f ` \<tau>s) seq i" for seq i
    by(induct rule: I.traversed.induct, auto intro: I'.traversed.intros)

  show ?thesis
  proof (intro exI conjI)
    from \<tau>s dom show "f ` \<tau>s \<subseteq> sharp_transitions_of P'" by auto
    have "traverse \<tau>s \<subseteq> I.traverse \<tau>s" apply(rule traverse_mono) using mono assms by auto
    also have "... \<subseteq> I'.traverse (f ` \<tau>s)" unfolding I'.traverse_def I.traverse_def using * by auto
    also have "... \<subseteq> traverse (f ` \<tau>s) O ?S\<^sup>*"
      apply(rule traverse_push2[OF change_fresh_push]) using dom assms by auto
    finally show "traverse \<tau>s \<subseteq> ...".
  qed (insert nonemp, auto)
qed

end

context
  fixes P P' :: "('f,'v,'t,'l sharp) lts" and x :: 'v and ty :: 't
  assumes fv: "fresh_variable_cond P P' x ty"
begin

lemma fresh_variable_addition:
  assumes "cooperation_SN P'"
  shows "cooperation_SN P"
  unfolding cooperation_SN_on_as_SN_on_traverse
proof(intro allI impI)
  fix \<tau>s assume \<tau>s: "\<tau>s \<subseteq> sharp_transitions_of P" "\<tau>s \<noteq> {}"

  let ?S = "change_fresh x ty"
  let ?F = "transitions_step (flat_transitions_of P)"
  let ?F' = "transitions_step (flat_transitions_of P')"

  from fresh_variable_condD(1)[OF fv]
  have [simp]: "initial_states P' = initial_states P" by (auto simp: initial_states_def)

  from sharp_traverse_simulate[OF fv \<tau>s] obtain \<tau>s'
  where \<tau>s': "\<tau>s' \<subseteq> sharp_transitions_of P'"
    and nonemp: "\<tau>s' \<noteq> {}"
    and sim: "traverse \<tau>s \<subseteq> traverse \<tau>s' O (change_fresh x ty)\<^sup>*" (is "?T \<subseteq> ?T' O _") by auto
  from \<tau>s' nonemp assms[unfolded cooperation_SN_on_as_SN_on_traverse]
  have "SN_on ?T' (?F'\<^sup>* `` initial_states P)" by auto
  from SN_on_Image_push[OF change_fresh_O_traverse this, OF fv] \<tau>s' SN_on_mono
  have SN: "SN_on ?T' ((?F'\<^sup>* O ?S\<^sup>*) `` initial_states P)" by (auto simp: relcomp_Image)
  note [simp] = rtrancl_U_push[OF change_fresh_O_flat_transitions[OF fv]]
  from flat_transitions_simulate[OF fv] have "?F\<^sup>* \<subseteq> (?F' O ?S\<^sup>*)\<^sup>*" by (intro rtrancl_mono,auto)
  also have "... \<subseteq> (?S \<union> ?F')\<^sup>*" by regexp
  also have "... = (?F')\<^sup>* O ?S\<^sup>*" by auto
  finally have "?F\<^sup>* \<subseteq> ?F'\<^sup>* O ?S\<^sup>*" by auto
  from SN_on_subset2[OF Image_mono[OF this] SN, OF subset_refl]
  have "SN_on ?T' (?F\<^sup>* `` initial_states P)" by auto
  from SN_on_O_push[OF change_fresh_O_traverse, OF fv _ this] \<tau>s'
  have "SN_on (?T' O ?S\<^sup>*) (?F\<^sup>* `` initial_states P)" by auto
  from SN_on_mono[OF this sim]
  show "SN_on ?T (?F\<^sup>* `` initial_states P)" by auto
qed

end

definition "definability_checker x ty \<phi> \<equiv> undefined"

lemma definability_checker:
  assumes "definability_checker x ty \<phi>"
  shows "\<exists>a \<in> type_values ty. \<alpha>((x,ty):=a) \<Turnstile> \<phi>"
sorry


definition "fresh_variable_checker P f x ty \<equiv>
  \<forall>(tr,\<tau>) \<in> set (transitions_impl P).
  definability_checker (Post x) ty (f tr) \<and>
  (Post x,ty) \<notin> vars_formula (transition.formula \<tau>) \<and>
  (Pre x, ty) \<notin> vars_formula (transition.formula \<tau>) \<and>
  (Pre x, ty) \<notin> vars_formula (f tr)"

end

lemma(in "theory") fresh_variable_checker:
  assumes ok: "fresh_variable_checker P f x ty"
  shows "fresh_variable_cond (lts_of P) (lts_of (refine_transition_formulas P f)) x ty"
    (is "fresh_variable_cond _ ?P x ty")
proof(unfold fresh_variable_cond_def, intro conjI ballI2 ballI)
  note ok = ok[unfolded fresh_variable_checker_def]
  from ok show "lts.initial ?P = lts.initial (lts_of P)" by auto
  let ?tf = transition.formula
  { fix \<tau>
    assume \<tau>: "\<tau> \<in> transitions (lts_of P)"
    then obtain tr where tr: "(tr,\<tau>) \<in> set (transitions_impl P)" by auto
    with ok
    have 1: "definability_checker (Post x) ty (f tr)"
     and 2: "(Post x, ty) \<notin> vars_formula (?tf \<tau>)"
     and 3: "(Pre x, ty) \<notin> vars_formula (f tr)" by auto
    show "\<exists>\<tau>'\<in>transitions ?P.
            source \<tau>' = source \<tau> \<and>
            target \<tau>' = target \<tau> \<and>
            (\<forall>\<alpha>. assignment \<alpha> \<longrightarrow> \<alpha> \<Turnstile> ?tf \<tau> \<longrightarrow> (\<exists>a \<in> type_values ty. \<alpha>((Post x, ty) := a) \<Turnstile> ?tf \<tau>'))"
    proof(rule bexI, intro conjI allI impI)
      let ?\<tau> = "Transition (source \<tau>) (target \<tau>) (?tf \<tau> \<and>\<^sub>f f tr)"
      from tr show "?\<tau> \<in> transitions ?P" by (auto simp: refine_transition_formulas_def)
      fix \<alpha>
      assume \<alpha>: "assignment \<alpha>"
         and 4: "\<alpha> \<Turnstile> ?tf \<tau>"
      from definability_checker[OF 1] obtain a
      where 5: "a \<in> local.type_values ty"
        and 6: "\<alpha>((Post x, ty) := a) \<Turnstile> f tr" by auto
      from 2 4 have "\<alpha>((Post x, ty) := a) \<Turnstile> ?tf \<tau>" by auto
      with 6 have "\<alpha>((Post x, ty) := a) \<Turnstile> transition.formula ?\<tau>" by auto
      with 5 show "\<exists>a \<in> local.type_values ty. \<alpha>((Post x, ty) := a) \<Turnstile> transition.formula ?\<tau>" by auto
    qed auto
  }
  { fix \<tau>' s t a
    assume \<tau>': "\<tau>' \<in> transitions ?P"
       and st: "(s,t) \<in> transition_step \<tau>'"
       and a: "a \<in> type_values ty"
    let ?s = "s |\<^sub>s (x,ty) := a"
    from \<tau>' obtain tr \<tau>
    where tr: "(tr,\<tau>) \<in> set (transitions_impl P)"
      and \<tau>': "\<tau>' = Transition (source \<tau>) (target \<tau>) (transition.formula \<tau> \<and>\<^sub>f f tr)"
      by (auto simp: refine_transition_formulas_def)
    with ok
    have 1: "definability_checker (Post x) ty (f tr)"
     and 2: "(Post x, ty) \<notin> vars_formula (?tf \<tau>)"
     and 3: "(Pre x, ty) \<notin> vars_formula (?tf \<tau>)"
     and 4: "(Pre x, ty) \<notin> vars_formula (f tr)" by auto
    let ?ppi = "pre_post_inter"
    let ?\<alpha> = "valuation s"
    let ?\<beta> = "valuation t"
    from \<tau>' st obtain \<gamma>
    where \<gamma>: "assignment \<gamma>" and "?ppi ?\<alpha> ?\<beta> \<gamma> \<Turnstile> transition.formula \<tau> \<and>\<^sub>f f tr" by auto
    then have 5: "?ppi ?\<alpha> ?\<beta> \<gamma> \<Turnstile> ?tf \<tau>" and 6: "?ppi ?\<alpha> ?\<beta> \<gamma> \<Turnstile> f tr" by auto
    from definability_checker[OF 1] obtain b
    where b: "b \<in> type_values ty"
      and *: "(?ppi (valuation ?s) ?\<beta> \<gamma>)((Post x, ty) := b) \<Turnstile> f tr" by blast
    let ?t = "t |\<^sub>s (x,ty) := b"
    from * have "?ppi (valuation ?s) (valuation ?t) \<gamma> \<Turnstile> f tr" by auto
    moreover
      from 5 have "(?ppi ?\<alpha> ?\<beta> \<gamma>)((Pre x, ty) := a) \<Turnstile> ?tf \<tau>"
        unfolding satisfies_with_fresh_var[OF 3] by auto
      then have "?ppi (valuation ?s) ?\<beta> \<gamma> \<Turnstile> ?tf \<tau>" by auto
      then have "(?ppi (valuation ?s) ?\<beta> \<gamma>)((Post x, ty) := b) \<Turnstile> ?tf \<tau>"
        unfolding satisfies_with_fresh_var[OF 2] by auto
      then have "?ppi (valuation ?s) (valuation ?t) \<gamma> \<Turnstile> ?tf \<tau>" by auto
    ultimately have "?ppi (valuation ?s) (valuation ?t) \<gamma> \<Turnstile> ?tf \<tau>'" by (auto simp: \<tau>')
    with \<gamma> a st b
    have "(?s,?t) \<in> transition_step \<tau>'" by(auto intro!: mem_transition_stepI[of _ _ \<gamma>] assignment_update)
    with b show "\<exists>b \<in> type_values ty. (?s, t |\<^sub>s (x, ty) := b) \<in> transition_step \<tau>'" by auto
  }
qed

lemma(in "theory") fresh_variable_addition_impl:
  assumes ok: "fresh_variable_checker P f x ty"
  shows "cooperation_SN (lts_of (refine_transition_formulas P f)) \<Longrightarrow> cooperation_SN (lts_of P)"
  using fresh_variable_addition[OF fresh_variable_checker[OF ok]]
  by auto


end