theory Invariant_Checker
imports 
  "$AFP/Check_Monad/Check_Monad"
  "~~/src/HOL/Library/RBT_Mapping"
  Formula
  LTS
begin

text \<open>In this theory we provide an invariant checker which establishes the 
  @{const pretheory.disj_invariant} property.\<close>


datatype ('f,'v,'t,'l,'n) art_edge = is_cover_node: Cover 'n | Children "(('f,'v,'t,'l) transition \<times> 'n) list"

record ('f,'v,'t,'l,'n) art =
  initial_node :: 'n
  nodes :: "'n set"
  edge :: "'n \<Rightarrow> ('f,'v,'t,'l,'n) art_edge"
  node_location :: "'n \<Rightarrow> 'l"
  node_invariant :: "'n \<Rightarrow> ('f,'v,'t) formula"

definition (in pretheory) transition_entailment ::
  "('f,'v,'t) formula \<Rightarrow> ('f,'v,'t,'l) transition \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> bool" where
  "transition_entailment \<phi> \<tau> \<psi> \<equiv>
   rename_vars Pre \<phi> \<and>\<^sub>f transition.formula \<tau> \<longrightarrow>\<^sub>f rename_vars Post \<psi>"

locale pre_art_body = "pretheory" T
  for T :: "('f,'t,'d) theory"
  and P :: "('f,'v,'t,'l) lts"
  and A :: "('f,'v,'t,'l,'n) art"
begin

definition "art \<equiv> \<forall>a \<in> nodes A. formula (node_invariant A a)"

lemma artD[dest!]: "art \<Longrightarrow> a \<in> nodes A \<Longrightarrow> formula (node_invariant A a)" by (auto simp:art_def)

definition "art_checker \<equiv> check art (shows ''ill-formed invariant'')"

lemma art_checker[intro]: "isOK art_checker \<Longrightarrow> art" unfolding art_checker_def by auto

definition "initial_cond \<equiv>
  initial_node A \<in> nodes A \<and>
  node_invariant A (initial_node A) = form_True"

definition "simulation_cond \<equiv>
  node_location A (initial_node A) = lts.initial P \<and>
  (\<forall> a children \<tau>.
    a \<in> nodes A
    \<longrightarrow> edge A a = Children children
    \<longrightarrow> \<tau> \<in> transitions P
    \<longrightarrow> source \<tau> = node_location A a
    \<longrightarrow> (\<exists> b \<in> nodes A. (\<tau>, b) \<in> set children \<and> node_location A b = target \<tau>))"

definition "cover_edges_cond \<equiv> \<forall> a b.
    a \<in> nodes A
    \<longrightarrow> edge A a = Cover b
    \<longrightarrow> b \<in> nodes A \<and>
      node_location A a = node_location A b \<and>
      \<not> is_cover_node (edge A b) \<and>
      node_invariant A a \<longrightarrow>\<^sub>f node_invariant A b"

definition "children_edges_cond \<equiv> \<forall> a children \<tau> b.
    a \<in> nodes A
    \<longrightarrow> edge A a = Children children
    \<longrightarrow> (\<tau>,b) \<in> set children
    \<longrightarrow> transition_entailment (node_invariant A a) \<tau> (node_invariant A b)"

definition "well_formed \<equiv>
  children_edges_cond \<and> cover_edges_cond \<and> initial_cond \<and> simulation_cond \<and> art"

fun corresponds_to :: "('v,'t,'d,'l) state \<Rightarrow> 'n \<Rightarrow> bool" where
  "corresponds_to (State \<alpha> l) a  = (node_location A a = l \<and> (\<alpha> \<Turnstile> node_invariant A a))"

(* LTS.pretheory.pre_post_inter_satisfies_rename_vars_pre
lemma pre_valuation:
  "(pre_post_inter \<alpha>1 \<alpha>2 \<beta> \<Turnstile> rename_vars Pre \<phi>) \<longleftrightarrow> (\<alpha>1 \<Turnstile> \<phi>)" (is "?l \<longleftrightarrow> ?r")
proof
  assume ?l
  from this[unfolded rename_vars_formula_def] have "\<And>cl. cl \<in> set \<phi> \<Longrightarrow> \<exists>lit \<in> set cl.
       eval_true (pre_post_inter \<alpha>1 \<alpha>2 \<beta>) (rename_vars Pre lit)" by fastforce
  then show ?r by auto
qed (intro satisfiesI, auto simp:rename_vars_formula_def)

lemma post_valuation:
  "(pre_post_inter \<alpha>1 \<alpha>2 \<beta> \<Turnstile> rename_vars Post \<phi>) \<longleftrightarrow> (\<alpha>2 \<Turnstile> \<phi>)" (is "?l \<longleftrightarrow> ?r")
proof
  assume ?l
  from this[unfolded rename_vars_formula_def] have "\<And>cl. cl \<in> set \<phi> \<Longrightarrow> \<exists>lit \<in> set cl.
       eval_true (pre_post_inter \<alpha>1 \<alpha>2 \<beta>) (rename_vars Post lit)" by fastforce
  then show ?r by auto
qed (intro satisfiesI, auto simp:rename_vars_formula_def) *)

lemma formula_node_invariant [intro]: "well_formed \<Longrightarrow> a \<in> nodes A \<Longrightarrow> formula (node_invariant A a)"
  unfolding well_formed_def art_def by auto

lemma well_formed_imp_disj_invariant:
  assumes well_formed
  and lts: "lts P"
  shows "disj_invariant_set P (\<lambda> loc. {node_invariant A a | a. a \<in> nodes A \<and> node_location A a = loc})"
  unfolding disj_invariant_set_def
proof (intro allI impI)
  fix \<alpha> loc
  assume reach: "State \<alpha> loc \<in> reachable_states P"
  note [intro] = formula_node_invariant[OF assms(1)]
  from reachable_state[OF reach] have \<alpha>: "assignment \<alpha>" by auto
  from reach obtain init
    where init: "init \<in> initial_states P"
      and steps: "(init, State \<alpha> loc) \<in> (transition P)^*" by auto
  from assms[unfolded well_formed_def]
  have ic: initial_cond 
   and sc: simulation_cond and cec: children_edges_cond
   and cc: cover_edges_cond and art: art by auto
  obtain l \<alpha>0 where "init = State \<alpha>0 l" by (cases init, auto)
  with init[unfolded initial_states_def]
  have "init = State \<alpha>0 (lts.initial P)" "assignment \<alpha>0" by auto
  with ic sc have ic: "initial_node A \<in> nodes A" "corresponds_to init (initial_node A)"
    by (auto simp: simulation_cond_def initial_cond_def)
  {
    fix s1 s2 a1
    assume t: "(s1,s2) \<in> transition P" and s1a1: "corresponds_to s1 a1" and a1: "a1 \<in> nodes A"
    obtain \<alpha>1 l1 where s1: "s1 = State \<alpha>1 l1" by (cases s1, auto)
    obtain \<alpha>2 l2 where s2: "s2 = State \<alpha>2 l2" by (cases s2, auto)
    from t[unfolded transition_def s1 s2] obtain \<tau> \<phi> \<beta>
    where trans: "\<tau> \<in> transitions P" "l1 = source \<tau> \<and> l2 = target \<tau> \<and> \<phi> = transition.formula \<tau>"
      and transf: "pre_post_inter \<alpha>1 \<alpha>2 \<beta> \<Turnstile> \<phi>"
      and \<alpha>1: "assignment \<alpha>1"
      and \<alpha>2: "assignment \<alpha>2"
      and \<beta>: "assignment \<beta>"
      by auto
    from lts trans have \<phi>: "formula \<phi>" unfolding lts_def by auto
    let ?\<alpha> = "(pre_post_inter \<alpha>1 \<alpha>2 \<beta>)"
    from \<alpha>1 \<alpha>2 \<beta> have ass: "assignment ?\<alpha>"
      by (auto simp: assignment_def, case_tac "x", auto)
    have "\<exists> a1 tas. corresponds_to s1 a1 \<and> edge A a1 = Children tas \<and> a1 \<in> nodes A"
    proof (cases "edge A a1")
      case (Children tas)
      with s1a1 a1 show ?thesis by blast
    next
      case (Cover b)
      from cc[unfolded cover_edges_cond_def, rule_format, OF a1 Cover] obtain tas where
        b: "b \<in> nodes A" and loc: "node_location A a1 = node_location A b"
        and edge: "edge A b = Children tas"
        and impl: "node_invariant A a1 \<longrightarrow>\<^sub>f node_invariant A b" by (cases "edge A b", auto)
      show ?thesis
      proof (rule exI[of _ b], rule exI[of _ tas], intro conjI)
        from s1a1[unfolded s1, simplified] loc
        have "node_location A b = l1" and "\<alpha>1 \<Turnstile> node_invariant A a1" by auto
        with impl a1 b \<alpha>1 show "corresponds_to s1 b" unfolding s1 by auto
      qed (insert b edge, auto)
    qed
    then obtain a1 tas where s1a1: "corresponds_to s1 a1"
      and edge: "edge A a1 = Children tas"
      and a1: "a1 \<in> nodes A" by blast
    from s1a1[unfolded s1] have l1: "l1 = node_location A a1"
      and \<alpha>1: "\<alpha>1 \<Turnstile> node_invariant A a1" by auto
    from sc[unfolded simulation_cond_def, THEN conjunct2,
      rule_format, OF a1 edge] trans l1 obtain b where
      b: "b \<in> nodes A" and mem: "(\<tau>, b) \<in> set tas" and l2: "node_location A b = l2" by auto
    
    have "\<alpha>2 \<Turnstile> node_invariant A b"
    proof(rule satisfies_rename_vars)
      from cec[unfolded children_edges_cond_def, rule_format, OF a1 edge mem]
      have "transition_entailment (node_invariant A a1) \<tau> (node_invariant A b)".
      from this[unfolded transition_entailment_def] trans(2)
      have *: "rename_vars Pre (node_invariant A a1) \<and>\<^sub>f \<phi> \<longrightarrow>\<^sub>f rename_vars Post (node_invariant A b)" by simp
      show "?\<alpha> \<Turnstile> ..."
        apply(intro impliesD[OF *] satisfies_andI formula_andI)
        unfolding pre_post_inter_satisfies_rename_vars_pre using \<alpha>1 transf ass a1 \<phi> b by auto
    qed auto
    with l2 have "corresponds_to s2 b" unfolding s2 by simp
    with b have "\<exists> a2 \<in> nodes A. corresponds_to s2 a2" by blast
  } note one_step_simulation = this
  {
    fix s1 s2 a1
    assume t: "(s1,s2) \<in> (transition P)^*" and mem: "a1 \<in> nodes A" and s1a1: "corresponds_to s1 a1"
    from t have "\<exists> a2 \<in> nodes A. corresponds_to s2 a2" 
    proof (induct rule: rtrancl_induct)
      case (step s2 s3)
      from step(3) obtain a2 where a2: "a2 \<in> nodes A" and s2a2: "corresponds_to s2 a2" by blast
      from one_step_simulation[OF step(2) s2a2 a2] show ?case .
    qed (insert s1a1 mem, auto)
  } note many_step_simulation = this
  from many_step_simulation[OF steps ic]
  obtain a_loc where a_loc: "a_loc \<in> nodes A" and err: "corresponds_to (State \<alpha> loc) a_loc" by auto
  then have "node_location A a_loc = loc \<and> (\<alpha> \<Turnstile> node_invariant A a_loc)" by auto
  with a_loc
  show "\<exists>\<phi>\<in>{node_invariant A a |a. a \<in> art.nodes A \<and> node_location A a = loc}. \<alpha> \<Turnstile> \<phi>" by auto
qed

end

record ('h,'l,'n,'tr) hinter =
  nodes :: "'n list"
  succ_trans_list :: "'l \<Rightarrow> 'tr list"
  cover_hints :: "'n \<Rightarrow> 'h list option"
  transition_hints :: "'n \<Rightarrow> 'h list list option"

hide_const (open) nodes

type_synonym ('n,'h) error_hints = "'n \<Rightarrow> 'h list option"

locale pre_art_checker = pretheory T
  for T :: "('f::show,'t::show,'d) theory" +
  fixes tc :: "('h::show,'f,'v::show,'t) theory_checker"
  and tc2 :: "('h,'f,'v trans_var,'t) theory_checker"

locale pre_art_checker_body =
  pre_art_checker T tc tc2 + pre_art_body T P A
  for T :: "('f::show,'t::show,'d) theory"
  and tc :: "('h::show,'f,'v::show,'t) theory_checker"
  and tc2
  and P :: "('f,'v,'t,'l::show) lts"
  and A :: "('f, 'v, 't, 'l, 'n::show) art" +
  fixes H :: "('h,'l,'n,('f,'v,'t,'l) transition) hinter"
begin

definition simulation_cond_checker :: "shows check" where
  "simulation_cond_checker \<equiv> do {
    check (node_location A (initial_node A) = lts.initial P)
      (shows ''initial node of ART does not correspond to initial node of program'');
    check_allm (\<lambda> a.
      case edge A a
      of Cover _ \<Rightarrow> succeed
      | Children children \<Rightarrow>
          let l = node_location A a in
          check_allm
            (\<lambda> \<tau>.
              check (\<exists> (\<tau>',b) \<in> set children. \<tau>' = \<tau> \<and> node_location A b = target \<tau> \<and> b \<in> art.nodes A)
            (shows ''could not find matching transition in ART for node '' o
             shows a o shows '' and transition ''(* o shows \<tau>*)))
            (succ_trans_list H l)
    ) (hinter.nodes H)
  } <+? (\<lambda> s. shows ''could not ensure simulation condition of ART\<newline>'' o s)"

definition cover_edges_cond_checker :: "shows check" where
  "cover_edges_cond_checker \<equiv> do {
     check_allm (\<lambda> a.
       case edge A a of Children _ \<Rightarrow> succeed
       | Cover b \<Rightarrow> do {
           check (b \<in> art.nodes A) (shows ''target node is not listed as art-node'');
           check (node_location A a = node_location A b) (shows ''node-locations differ'');
           check (\<not> is_cover_node (edge A b)) (shows ''target node must not have cover edge'');
           case cover_hints H a of None \<Rightarrow> error (shows ''there is no cover hint present'')
           | Some hints \<Rightarrow> pre_theory_checker.formula_checker tc hints (node_invariant A a) (node_invariant A b)
       } <+? (\<lambda> s. shows ''problem in checking cover edge of node '' o shows a o shows_nl o s)
     ) (hinter.nodes H)
   } <+? (\<lambda> s. shows ''could not ensure cover edge condition of ART\<newline>'' o s)" 

definition initial_cond_checker :: "shows check" where
  "initial_cond_checker \<equiv> do {
    let init = initial_node A;
    check (init \<in> art.nodes A)
      (shows ''initial node of A ('' o shows init o shows '') is not mentioned as node of ART''
      );
    check (node_invariant A init = form_True)
     (shows ''the node invariant for the initial ART node ('' o shows init o shows '') must be TRUE'')
  }"

definition children_edges_cond_checker :: "shows check" where
  "children_edges_cond_checker \<equiv> do {
     check_allm (\<lambda> a.
       case edge A a of Cover _ \<Rightarrow> succeed
       | Children children \<Rightarrow> do {
           case transition_hints H a of None \<Rightarrow> error (shows ''there is no transition hint present'')
           | Some hintss \<Rightarrow> do {
               check (length hintss = length children)
                 (shows ''the number of hints differs from the number of children'');
               check_allm (\<lambda> ((\<tau>,b),hints).
                 pre_theory_checker.formula_checker tc2 hints
                   (rename_vars_formula Pre (node_invariant A a) \<and>\<^sub>f transition.formula \<tau>)
                   (rename_vars_formula Post (node_invariant A b)) <+?
                 (\<lambda> s. shows ''problem in checking transition '' o shows a o shows '' --'' o
                   (* shows \<tau> o*) shows ''--> '' o shows b o shows_nl o s)
               ) (zip children hintss)
             }
       } <+? (\<lambda> s. shows ''problem in checking transitions of node '' o shows a o shows_nl o s)
     ) (hinter.nodes H)
   } <+? (\<lambda> s. shows ''could not ensure transition edge condition of ART\<newline>'' o s)" 

definition art_invariant_checker_main  :: "shows check" where
  "art_invariant_checker_main \<equiv> do {
      art_checker;
      initial_cond_checker;
      simulation_cond_checker;
      cover_edges_cond_checker;
      children_edges_cond_checker
   } <+? (\<lambda> s. shows ''could not ensure invariants in ART:\<newline>'' o s)"

definition error_cond_checker :: "'l \<Rightarrow> ('n,'h) error_hints \<Rightarrow> shows check" where
  "error_cond_checker err hints \<equiv> do {
     check_allm (\<lambda> a.
       if node_location A a = err then
         case hints a of None \<Rightarrow>
           error (shows ''could not find hint for checking error condition of node '' o shows a)
         | Some hints \<Rightarrow> pre_theory_checker.formula_checker tc hints (node_invariant A a) form_False
       else succeed)
       (hinter.nodes H)
   } <+? (\<lambda> s. shows ''could not ensure error condition of ART\<newline>'' o s)"

definition art_safety_checker_main where
  "art_safety_checker_main err hints \<equiv> do {
      art_invariant_checker_main;
      error_cond_checker err hints
   } <+? (\<lambda> s. shows ''could not ensure safety property via ART:\<newline>'' o s)"

end

locale art_checker =
  pre_art_checker T tc tc2 +
  tc: theory_checker T tc +
  tc2: theory_checker T tc2
  for T :: "('f::show,'t::show,'d) theory"
  and tc :: "('h::show,'f,'v::show,'t) theory_checker"
  and tc2

locale art_checker_body =
  art_checker T tc tc2 +
  pre_art_checker_body T tc tc2 P A H
  for T :: "('f::show,'t::show,'d) theory"
  and tc :: "('h::show,'f,'v::show,'t) theory_checker"
  and tc2
  and P :: "('f,'v,'t,'l::show) lts"
  and A
  and H +
  assumes ans: "set (hinter.nodes H) = art.nodes A"
  and succ_trans_list: "\<And> l. set (succ_trans_list H l) = { \<tau> \<in> transitions P. source \<tau> = l }"
begin

lemma initial_cond_checker[iff]: "isOK(initial_cond_checker) = initial_cond"
  unfolding initial_cond_def initial_cond_checker_def by auto

lemma cover_edges_cond_checker[dest]:
  assumes ok: "isOK(cover_edges_cond_checker)"
    shows "cover_edges_cond"
  unfolding cover_edges_cond_def
proof (intro allI impI conjI)
  note * = ok[unfolded cover_edges_cond_checker_def,simplified, unfolded ans]
  note [split] = art_edge.splits option.splits
  fix a b assume **: "a \<in> nodes A" "edge A a = Cover b"
  from * ** show "node_location A a = node_location A b" "b \<in> nodes A" "\<not> is_cover_node (edge A b)" by auto
  from *[rule_format,OF **(1), unfolded **(2),simplified]
  obtain hints where "isOK (tc.formula_checker hints (node_invariant A a) (node_invariant A b))" by auto
  then show "node_invariant A a \<longrightarrow>\<^sub>f node_invariant A b" by (intro tc.formula_checker)
qed

lemma simulation_cond_checker[dest]: assumes "isOK(simulation_cond_checker)"
  shows simulation_cond
proof -
  note ok = assms[unfolded simulation_cond_checker_def, simplified, unfolded Let_def ans]
  show ?thesis
    unfolding simulation_cond_def
  proof (intro conjI allI impI)
    fix a tas \<tau>
    assume a: "a \<in> art.nodes A"
    assume *: "edge A a = Children tas" "\<tau> \<in> transitions P" "source \<tau> = node_location A a"
    from conjunct2[OF ok, rule_format, OF a] *
    show "\<exists>b\<in>art.nodes A. (\<tau>, b) \<in> set tas \<and> node_location A b = target \<tau>" by (auto simp: succ_trans_list)
  qed (insert ok, auto)
qed

lemma children_edges_cond_checker[dest]:
  assumes ok: "isOK(children_edges_cond_checker)"
  shows "children_edges_cond"
  unfolding children_edges_cond_def
proof (intro allI impI)
  fix a children \<tau> b
  assume a: "a \<in> art.nodes A" 
    and edge: "edge A a = Children children" and mem: "(\<tau>, b) \<in> set children"
  note ok = ok[unfolded children_edges_cond_checker_def, simplified, unfolded ans, rule_format, OF a, 
    unfolded edge, simplified]
  from ok obtain hints where tr: "transition_hints H a = Some hints" and len: "length hints = length children"
    by (auto split: option.splits)
  from mem len obtain h where mem: "((\<tau>,b),h) \<in> set (zip children hints)" unfolding set_zip
    by (auto simp: set_conv_nth)
  from ok[unfolded tr, simplified,THEN conjunct2,rule_format,OF this]
  show "transition_entailment (node_invariant A a) \<tau> (node_invariant A b)"
    unfolding transition_entailment_def
    by(intro tc2.formula_checker,auto)
qed

lemma art_invariant_checker_main[dest]: "isOK(art_invariant_checker_main) \<Longrightarrow> well_formed"
unfolding well_formed_def art_invariant_checker_main_def by auto

end

datatype ('n,'tr,'h) art_edge_impl =
  Cover_Edge 'n 'h
| Children_Edge "('tr \<times> 'n \<times> 'h) list"

datatype ('f,'v,'t,'l,'n,'tr,'h) art_node_impl = Art_Node
  (name: 'n) (invariant: "('f,'v,'t) formula") (location: 'l) (edge: "('n,'tr,'h) art_edge_impl")

fun is_cover_node :: "('f,'v,'t,'l,'n,'tr,'h) art_node_impl \<Rightarrow> bool" where
  "is_cover_node (Art_Node _ _ _ (Cover_Edge _ _)) = True"
| "is_cover_node (Art_Node _ _ _ (Children_Edge _)) = False"

record ('f,'v,'t,'l,'n,'tr,'h) art_impl =
  initial_node :: 'n
  nodes :: "('f,'v,'t,'l,'n,'tr,'h) art_node_impl list"
  error_hints :: "('n \<times> 'h) list"

hide_const (open) nodes

definition art_nodes :: "('f,'v,'t,'l,'n,'tr,'h) art_impl \<Rightarrow> 'n list" where
  "art_nodes Ai = map name (art_impl.nodes Ai)"

context
  fixes Pi :: "('f,'v,'t,'l,'tr) lts_impl"
begin
fun art_edge_of :: "('n,'tr,'h) art_edge_impl \<Rightarrow> ('f,'v,'t,'l,'n) art_edge" where
  "art_edge_of (Cover_Edge an _) = Cover an"
| "art_edge_of (Children_Edge ans) = Children (map (\<lambda> (t,a,h). (transition_of Pi t, a)) ans)"

definition art_of :: "('f,'v,'t,'l,'n,'tr,'h) art_impl \<Rightarrow> ('f,'v,'t,'l,'n) art" where
  "art_of Ai \<equiv>
    let ans = art_impl.nodes Ai in \<lparr>
      art.initial_node = art_impl.initial_node Ai,
      art.nodes = set (art_nodes Ai),
      art.edge = (the o map_of (map (\<lambda> a. (name a, art_edge_of (edge a))) ans)),
      node_location = (the o map_of (map (\<lambda> a. (name a, location a)) ans)),
      node_invariant = (the o map_of (map (\<lambda> a. (name a, invariant a)) ans))
    \<rparr>"

lemma node_invariant_art_of[simp] :
  "node_invariant (art_of A) = the o map_of (map (\<lambda> a. (name a, invariant a)) (art_impl.nodes A))"
  by (simp add: art_of_def Let_def)

lemma nodes_art_of[simp]: "art.nodes (art_of Ai) = set (art_nodes Ai)" by (simp add: art_of_def Let_def)

definition error_hints :: "('f,'v,'t,'l,'n,'tr,'h) art_impl \<Rightarrow> 'n \<Rightarrow> 'h option" where
  "error_hints Ai = map_of (art_impl.error_hints Ai)"

definition cover_hints :: "('f,'v,'t,'l,'n,'tr,'h) art_impl \<Rightarrow> 'n \<Rightarrow> 'h option" where
  "cover_hints Ai =
    map_of (map (\<lambda> a. (name a, case edge a of Cover_Edge _ h \<Rightarrow> h))
      (filter is_cover_node (art_impl.nodes Ai)))"

definition transition_hints :: "('f,'v,'t,'l,'n,'tr,'h) art_impl \<Rightarrow> 'n \<Rightarrow> 'h list option" where
  "transition_hints Ai =
    map_of (map (\<lambda> a. (name a, case edge a of Children_Edge cs \<Rightarrow> map (\<lambda> (_,_,h). h) cs))
      (filter (\<lambda> a. \<not> is_cover_node a) (art_impl.nodes Ai)))"

end

context pre_art_checker begin

definition "make_hinter Pi Ai \<equiv> \<lparr>
  hinter.nodes = art_nodes Ai,
  succ_trans_list = succ_transitions Pi,
  cover_hints = cover_hints Ai,
  transition_hints = transition_hints Ai
\<rparr>"

definition unique_names where
"unique_names Ai \<equiv> distinct (map name (art_impl.nodes Ai))"

definition unique_names_checker
where
  "unique_names_checker Ai \<equiv>
    check (card (set (art_nodes Ai)) = length (art_impl.nodes Ai))
          (shows ''Nodes in art graph must have unique names'')"

lemma unique_names_checker:
  assumes "isOK(unique_names_checker Ai)"
  shows "unique_names Ai"
proof -
  have "card (set (map name (art_impl.nodes Ai))) = length (map name (art_impl.nodes Ai))"
    using assms unfolding unique_names_def unique_names_checker_def art_nodes_def by auto
  from this[folded length_remdups_card,unfolded length_remdups_eq remdups_id_iff_distinct]
  show ?thesis unfolding unique_names_def.
qed

definition art_invariant_checker ::
  "('f::show,'v::show,'t::show,'l::show,'n::show,'tr,'h list) art_impl
  \<Rightarrow> ('f,'v,'t,'l,'tr) lts_impl \<Rightarrow> shows + ('l \<Rightarrow> ('f, 'v, 't) formula list)"
where
  "art_invariant_checker Ai Pi \<equiv>
   let P = lts_of Pi;
       A = art_of Pi Ai;
       H = make_hinter Pi Ai;
       eH = error_hints Ai
   in (pre_art_checker_body.art_invariant_checker_main T tc tc2 P A H \<then>
       return (\<lambda> loc. map (node_invariant A) (filter (\<lambda> v. node_location A v = loc) (art_nodes Ai)))
      ) <+?
   (\<lambda> s. shows ''could not ensure that the invariants hold via ART:\<newline>'' o s)"

abbreviation post :: "('f,'v,'t) formula \<Rightarrow> ('f,'v trans_var,'t) formula" where
  "post f \<equiv> rename_vars Post f"


fun translate_f
where
  "translate_f is_cover f 0 = f 0" |
  "translate_f is_cover f (Suc n)
    = (if is_cover (f 1) then translate_f is_cover (shift f 2) n else translate_f is_cover (shift f 1) n)"

lemma translate_f_skips:
  assumes "\<And> n. is_cover (f n) \<Longrightarrow> \<not> is_cover (f (Suc n))"
  and "\<not> is_cover (f 0)"
  shows "\<not> is_cover (translate_f is_cover f n)"
using assms by (induct n arbitrary: f;cases "is_cover (f 1)";auto)

definition get_node where
  "get_node Ai = the o map_of (map (\<lambda> a. (name a, a)) (art_impl.nodes Ai))"

abbreviation node_exists where
  "node_exists Ai x \<equiv> x \<in> name ` set (art_impl.nodes Ai)"

definition is_cover_state where
  "is_cover_state Ai = is_cover_node \<circ> get_node Ai o state.location"

lemma unique_names_eq:
  assumes "unique_names Ai"
          "name node = name a"
          "node \<in> set (art_impl.nodes Ai)"
          "a \<in> set (art_impl.nodes Ai)"
  shows "node = a"
using distinct_map_eq[OF assms(1)[unfolded unique_names_def] assms(2-4)].

abbreviation satisfies_invariant where
  "satisfies_invariant Ai x' \<equiv> (assignment (valuation x') \<and> node_exists Ai (state.location x') \<and> (valuation x') \<Turnstile> (invariant (get_node Ai (state.location x'))))"

lemma map_of_helper: (* Makes a great SIMP rule, but really slows down simp/force *)
  assumes "map_of (map (\<lambda>a. (f1 a, a)) lst) l = Some v"
  shows "map_of (map (\<lambda>a. (f1 a, f a)) lst) l = Some (f v)"
using assms by(induct lst;auto)

lemma get_node :
  assumes "l \<in> name ` set (art_impl.nodes Ai)"
  shows "get_node Ai l \<in> set (art_impl.nodes Ai)" (is ?t1)
        "name (get_node Ai l) = l" (is ?t2)
        "map_of (map (\<lambda>a. (name a, a)) (art_impl.nodes Ai)) l = Some (get_node Ai l)" (is ?t3)
proof -
  from assms weak_map_of_SomeI obtain a where a:"name a = l" "a \<in> set (art_impl.nodes Ai)" by auto
  hence ins:"(name a,a) \<in> set (map (\<lambda> a. (name a, a)) (art_impl.nodes Ai))" by auto
  obtain x where x:"map_of (map (\<lambda>a. (name a, a)) (art_impl.nodes Ai)) (name a) = Some x"
           using weak_map_of_SomeI[OF ins] by auto
  hence x_is:"x = get_node Ai l" unfolding get_node_def o_def a by simp
  from map_of_SomeD[OF x,unfolded a x_is] show ?t1 ?t2 by auto
  from x[unfolded x_is a] show ?t3 by auto
qed

lemma is_cover_state:
  assumes "node_exists Ai (state.location x)" "is_cover_state Ai x"
  shows "\<exists>v. map_of (map (\<lambda>a. (name a, art_edge_of Pi (art_node_impl.edge a)))
                                       (art_impl.nodes Ai)) (state.location x) = Some (Cover v)
              \<and> (\<exists> h. art_node_impl.edge (get_node Ai (state.location x)) = Cover_Edge v h)"
proof -
  have [simp]:"map_of (map (\<lambda>a. (name a, a)) (art_impl.nodes Ai)) (state.location x) =
        Some (get_node Ai (state.location x))"
    using get_node(3)[OF assms(1)].
  from assms(2)[unfolded is_cover_state_def,simplified] show ?thesis
  by (cases "(get_node Ai (state.location x))";cases "edge (get_node Ai (state.location x))"
     ;auto simp:map_of_helper)+
qed
lemma no_cover_state:
  assumes "node_exists Ai (state.location x)" "\<not> is_cover_state Ai x"
  shows "\<exists>v. map_of (map (\<lambda>a. (name a, art_edge_of Pi (art_node_impl.edge a)))
                                       (art_impl.nodes Ai)) (state.location x) = Some (Children v)"
proof -
  have [simp]:"map_of (map (\<lambda>a. (name a, a)) (art_impl.nodes Ai)) (state.location x) =
        Some (get_node Ai (state.location x))"
    using get_node(3)[OF assms(1)].
  from assms(2)[unfolded is_cover_state_def,simplified] show ?thesis
  by (cases "(get_node Ai (state.location x))";cases "edge (get_node Ai (state.location x))"
     ;auto simp:map_of_helper)
qed

definition lts_transitions_of_art
 :: "('f, 'v, 't, 'l, 'tr) lts_impl \<Rightarrow> ('f, 'v, 't, 'l, 'n, 'tr, 'hintList) art_node_impl \<Rightarrow> ('f,'v,'t,'n,'tr option) transitions_impl"
where
  "lts_transitions_of_art Pi a = (case (edge a) of
        (Cover_Edge v _) \<Rightarrow> [(None,Transition (name a) v (post (invariant a)))] |
        (Children_Edge cs) \<Rightarrow> map (\<lambda> (nm,an,_). (Some nm,Transition (name a) an (transition.formula (transition_of Pi nm)))) cs)"

fun renumber_rename :: "int \<Rightarrow> ('b option \<times> 'c) list \<Rightarrow> (('b + int) \<times> 'c) list" 
where
"renumber_rename i (Cons (a,v) rs)
  = Cons (case a of None \<Rightarrow> Inr i | Some l \<Rightarrow> Inl l,v) (renumber_rename (i+1) rs)" |
"renumber_rename i Nil = Nil"

lemma snd_rename [simp]:
  shows "map snd (renumber_rename n xs) = map snd xs"
by(induct xs arbitrary:n,auto)

lemma snd_set_rename [simp]:
  shows "snd ` set (renumber_rename n xs) = snd ` set xs"
using set_map[of snd xs, folded snd_rename[of n], unfolded set_map].

lemma lts_termination_renumber_rename:
  shows "lts_termination (lts_of (Lts_Impl i (renumber_rename n xs))) = lts_termination (lts_of (Lts_Impl i xs))"
by auto

definition relabel_transition
where "relabel_transition f t = Transition (f (source t)) (f (target t)) (transition.formula t)"

definition lts_impl_of_art ::
  "('f,'v,'t,'l,'tr) lts_impl \<Rightarrow> ('f,'v,'t,'l,'n,'tr,'hintList) art_impl
  \<Rightarrow> ('f,'v,'t,'n,'tr + int) lts_impl"
 (* TODO: change resulting datatype s.t. transfer-labels are unique again:
          use "(art_node \<times> 'tr) + int" for instance, or part of the hintList? *)
where
  "lts_impl_of_art Pi Ai
    = (Lts_Impl (initial_node Ai) (renumber_rename 0
                                  (concat (map (lts_transitions_of_art Pi) (art_impl.nodes Ai)))))"

end

context art_checker begin

lemma disj_invariant_checker:
  shows "disj_invariant_checker (art_invariant_checker Ai)"
unfolding disj_invariant_checker_def
proof(clarify)
  fix Pi :: "('f::show,'v::show,'t::show,'a,'b) lts_impl"
  assume lts:"lts (lts_of Pi)"
  hence body:"art_checker_body T tc tc2 (lts_of Pi) (art_of Pi Ai) (make_hinter Pi Ai)"
    by(unfold_locales,simp_all add: make_hinter_def)
  show "case art_invariant_checker Ai Pi of Inl x \<Rightarrow> True
         | Inr I \<Rightarrow> disj_invariant (lts_of Pi) I \<and> (\<forall>l. list_all local.formula (I l))"
    proof(cases "art_invariant_checker Ai Pi")
      case (Inr b)
        hence wf:"pre_art_body.well_formed T (lts_of Pi) (art_of Pi Ai)"
          using art_checker_body.art_invariant_checker_main[OF body]
          unfolding art_checker_body.art_invariant_checker_main art_invariant_checker_def Let_def
          by auto
        have b:"b = (\<lambda> loc. map (node_invariant (art_of Pi Ai))
                             (filter (\<lambda> v. node_location (art_of Pi Ai) v = loc) (art_nodes Ai)))"
          using Inr unfolding art_invariant_checker_def Let_def by auto
        have "\<And>a. a\<in>set (art_nodes Ai) \<Longrightarrow>
                   formula (the (map_of (map (\<lambda>a. (name a, invariant a)) (art_impl.nodes Ai)) a))"
          using wf unfolding pre_art_body.well_formed_def pre_art_body.art_def by simp
        hence laf:"(\<forall>l. list_all formula
          (map ((the \<circ>\<circ> map_of) (map (\<lambda>a. (name a, invariant a)) (art_impl.nodes Ai)))
            [v\<leftarrow>art_nodes Ai . node_location (art_of Pi Ai) v = l]))" 
          by (simp add: imageE pred_list_def)
        show ?thesis using pre_art_body.well_formed_imp_disj_invariant[OF wf lts]
          unfolding disj_invariant_def disj_invariant_set_def Inr b
          by(simp del:disj_invariant_is_disj_invariant_set add:laf) blast
    qed(auto)
qed

end

context art_checker begin

lemma path_simulation:
  assumes step:"(x,y) \<in> transition (lts_of Pi)"
     and  sim:"pre_art_body.simulation_cond (lts_of Pi) (art_of Pi Ai)"
              "pre_art_body.children_edges_cond T (art_of Pi Ai)"
     and  cov:"pre_art_body.cover_edges_cond T (art_of Pi Ai)"
     and  art:"pre_art_body.art T (art_of Pi Ai)"
     and  lts:"lts (lts_of Pi)"
     and  x_is_x:"relabel_state (location \<circ> get_node Ai) x' = x"
     and  sat:"satisfies_invariant Ai x'"
  shows "\<exists> y'. (x',y') \<in> (transition (lts_of (lts_impl_of_art Pi Ai)))^+
             \<and> (relabel_state (location \<circ> get_node Ai) y' = y
             \<and> satisfies_invariant Ai y')" (is "\<exists> y'. ?trancl x' y' \<and> ?x_is_x y' y \<and> ?sat y'")
proof -
  let ?x' = "(get_node Ai (state.location x'))"
  let ?A = "art_of Pi Ai"
  let ?path = "path (transition (lts_of (lts_impl_of_art Pi Ai)))"
  have x_in_Ai:"node_exists Ai (state.location x')" using sat by auto
  obtain \<tau> where \<tau>:"\<tau> \<in> transitions (lts_of Pi)" "(x,y) \<in> transition_step \<tau>"
    using step unfolding transition_def by auto
  then obtain \<gamma> where \<gamma>: "state x" "state y" "assignment \<gamma>" "source \<tau> = state.location x"
                         "target \<tau> = state.location y"
                         "\<delta> (valuation x) (valuation y) \<gamma> \<Turnstile> transition.formula \<tau>"
    unfolding transition_step_def by auto
  from \<tau> lts[unfolded lts_def] have \<tau>_formula: "formula (transition.formula \<tau>)" by auto
  have xinnodes:"state.location x' \<in> nodes ?A"
    unfolding art_of_def art_nodes_def Let_def using x_in_Ai by auto
  (* WLOG, we can assume that x is not a cover node *)
  have "\<exists> x1 i1. ?path x' i1 x1 \<and> ?x_is_x x1 x \<and> ?sat x1 \<and> \<not> is_cover_state Ai x1"
  proof(cases "is_cover_state Ai x'")
    case True
     obtain v h where v:"art.edge (art_of Pi Ai) (state.location x') = Cover v"
               "art_node_impl.edge (get_node Ai (state.location x')) = Cover_Edge v h"
        using is_cover_state[OF x_in_Ai True,of Pi] unfolding art_of_def Let_def o_def by auto
     let ?x1 = "State (state.valuation x') v"
     have res:"v \<in> nodes (art_of Pi Ai)"
          "node_location (art_of Pi Ai) (state.location x') = node_location (art_of Pi Ai) v"
          "\<not> art_edge.is_cover_node (art.edge (art_of Pi Ai) v)"
          "node_invariant (art_of Pi Ai) (state.location x') \<longrightarrow>\<^sub>f node_invariant (art_of Pi Ai) v"
          using cov[unfolded pre_art_body.cover_edges_cond_def,rule_format,OF xinnodes v(1)] by auto
     hence v_loc:"v \<in> name ` set (art_impl.nodes Ai)" unfolding art_of_def Let_def art_nodes_def by auto
     {fix a
       have "art_edge.is_cover_node (art_edge_of Pi (art_node_impl.edge a)) = is_cover_node a"
       by(cases "art_node_impl.edge a";cases a;auto)
     } note [simp] = this
     have nc:"\<not> is_cover_state Ai ?x1"
       using map_of_helper[OF get_node(3)[OF v_loc],of "art_edge_of Pi \<circ> art_node_impl.edge"]
       res(3) unfolding is_cover_state_def art_of_def Let_def by auto
     have newl:"?x_is_x ?x1 x"
       using x_is_x res(2) get_node(3)[OF v_loc] get_node(3)[OF x_in_Ai]
       by (auto simp:map_of_helper art_of_def Let_def)
     have frml:"formula (node_invariant (art_of Pi Ai) (state.location x'))"
               "formula (node_invariant (art_of Pi Ai) v)"
       using art[unfolded pre_art_body.art_def,rule_format,OF xinnodes]
             art[unfolded pre_art_body.art_def,rule_format,OF res(1)] by auto
     have assg:"assignment (state.valuation x')"
               "valuation x' \<Turnstile> node_invariant (art_of Pi Ai) (state.location x')"
       using sat get_node[OF v_loc] get_node[OF x_in_Ai] by (auto simp:map_of_helper)
     have newsat:"satisfies_invariant Ai ?x1"
       using get_node[OF v_loc] res(4)[unfolded implies_def,rule_format,OF frml assg] assg(1) v_loc
       by (auto simp:map_of_helper)
     have path:"path (transition (lts_of (lts_impl_of_art Pi Ai))) x' [x'] ?x1"
       proof(rule path_prepend[OF _ path0])
       (* instantiate gamma with (valuation x') *)
       have "\<exists>\<gamma>. assignment \<gamma> \<and> 
          name (get_node Ai (state.location x')) = state.location x' \<and> \<delta> (valuation x') (valuation x') \<gamma> \<Turnstile> post (invariant (get_node Ai (state.location x')))" 
       using assg sat get_node[OF x_in_Ai] by auto
       (* instantiate z with the result of lts_transitions_of_art *)
       hence "\<exists> z. z\<in>set (lts_transitions_of_art Pi (get_node Ai (state.location x')))
                 \<and> (\<exists>\<gamma>. assignment \<gamma> \<and> source (snd z) = state.location x'
                 \<and> target (snd z) = v
                 \<and> \<delta> (valuation x') (valuation x') \<gamma> \<Turnstile> transition.formula (snd z))"
         unfolding lts_transitions_of_art_def v by(simp)
       hence "\<exists>x\<in>set (art_impl.nodes Ai).
         \<exists>x\<in>set (lts_transitions_of_art Pi x).
          \<exists>\<gamma>. assignment \<gamma> \<and> source (snd x) = state.location x' \<and> target (snd x) = v \<and> \<delta> (valuation x') (valuation x') \<gamma> \<Turnstile> transition.formula (snd x)"
         using get_node(1)[OF x_in_Ai] by blast
       thus "(x', State (valuation x') v) \<in> transition (lts_of (lts_impl_of_art Pi Ai))"
            unfolding transition_def lts_impl_of_art_def by(auto simp: transition_step_def assg(1))
     qed
     from nc newl newsat path show ?thesis by auto next
    case False 
      have "?path x' [] x'" by auto thus ?thesis using False x_is_x sat by blast
  qed
  then obtain x1 i1 where x1: "?path x' i1 x1" "?x_is_x x1 x" "?sat x1" "\<not> is_cover_state Ai x1"
    by auto
  have x1_in_Ai:"node_exists Ai (state.location x1)" using x1 by auto
  hence l_isnode:"state.location x1 \<in> nodes (art_of Pi Ai)"
    by (auto simp:art_nodes_def)
  obtain v where v:"art.edge (art_of Pi Ai) (state.location x1) = Children v"
        using no_cover_state[OF x1_in_Ai x1(4),of Pi]
        unfolding art_of_def Let_def by auto
  then obtain lst where lst:
        "art_node_impl.edge (get_node Ai (state.location x1)) = Children_Edge lst"
        "v = map (\<lambda>(t, a, h). (transition_of Pi t, a)) lst"
        unfolding art_of_def Let_def o_def
        by (cases "art_node_impl.edge (get_node Ai (state.location x1))"
           ;auto simp:map_of_helper get_node(3)[OF x1_in_Ai])
  have "source \<tau> = node_location (art_of Pi Ai) (state.location x1)"
    using \<tau>(2) x1(2)
    by(auto simp:map_of_helper get_node(3)[OF x1_in_Ai]
       transition_step_def art_of_def Let_def)
  hence "(\<exists>b\<in>nodes (art_of Pi Ai). (\<tau>, b) \<in> set v \<and> node_location (art_of Pi Ai) b = target \<tau>)"
    using sim(1)[unfolded pre_art_body.simulation_cond_def] l_isnode v \<tau>
    by simp
  then obtain b where b:"b\<in>nodes (art_of Pi Ai)" "(\<tau>, b) \<in> set v"
                      "node_location (art_of Pi Ai) b = target \<tau>" by auto
  hence node_exists:"node_exists Ai b" unfolding art_of_def Let_def art_nodes_def by auto
  from \<tau>[unfolded transition_step_def] obtain \<gamma> where \<gamma>:
    "state x" "state y" "assignment \<gamma>" "source \<tau> = state.location x" "target \<tau> = state.location y"
    "\<delta> (valuation x) (valuation y) \<gamma> \<Turnstile> transition.formula \<tau>" by auto
  then have y:"node_location (art_of Pi Ai) b = state.location y" using b by auto
  have ea:"\<exists>\<gamma>. assignment \<gamma> \<and> \<delta> (valuation x1) (valuation y) \<gamma> \<Turnstile> transition.formula \<tau>"
    using \<gamma> x1(2) by auto
  let ?y = "State (state.valuation y) b"
  let ?d = "\<delta> (valuation x) (valuation y) \<gamma>"
  let ?trans = "Transition (state.location x1) b (transition.formula \<tau>)"
  have y_is_y:"?x_is_x ?y y" using y
    by(cases y;auto simp:art_of_def Let_def
                         map_of_helper get_node(3)[OF node_exists])
  have assg_d:"assignment ?d" using \<gamma> by auto
  obtain t h where x:"\<tau> = (transition_of Pi t)" 
                     "(t, b, h) \<in> set lst"
    using b(2)[unfolded lst] by auto
  have "?trans
    \<in> snd ` (\<lambda>(nm, an, _).
                 (Some nm,
                  Transition (name (get_node Ai (state.location x1))) an (transition.formula (transition_of Pi nm)))) `
             set lst"
    using x get_node[OF x1_in_Ai] by force 
  hence "?trans \<in> snd ` set (lts_transitions_of_art Pi (get_node Ai (state.location x1)))"
     unfolding lts_transitions_of_art_def lst by simp
  then obtain p where p:"p\<in>set (lts_transitions_of_art Pi (get_node Ai (state.location x1)))"
                        "snd p = ?trans" by auto
  hence "\<exists>\<gamma>. assignment \<gamma> \<and> source (snd p) = state.location x1 \<and> target (snd p) = b \<and>
             \<delta> (valuation x1) (valuation y) \<gamma> \<Turnstile> transition.formula (snd p)"
     using ea by auto
  hence "\<exists>x\<in>set (art_impl.nodes Ai).
         \<exists>x\<in>set (lts_transitions_of_art Pi x).
          \<exists>\<gamma>. assignment \<gamma> \<and> source (snd x) = state.location x1 \<and> target (snd x) = b \<and> \<delta> (valuation x1) (valuation y) \<gamma> \<Turnstile> transition.formula (snd x)"
     using get_node(1)[OF x1_in_Ai] p by blast
  from \<gamma>(1,2) x1(2) this
  have "(x1, State (valuation y) b) \<in> transition (lts_of (lts_impl_of_art Pi Ai))"
    unfolding transition_def by(auto simp:lts_impl_of_art_def transition_step_def)
  hence path:"?path x' (i1@Cons x1 Nil) ?y" by(rule path_conc[OF x1(1) path_prepend[OF _ path0]])
  find_theorems "_  \<in> set v"
  let ?pre = "rename_vars Pre (node_invariant (art_of Pi Ai) (state.location x1))"
  let ?post = "rename_vars Post (node_invariant (art_of Pi Ai) b)"
  have "valuation x1 \<Turnstile> invariant (get_node Ai (state.location x1))" using x1(3) by auto
  hence "?d \<Turnstile> ?pre" using x1(2)
    by (auto simp:map_of_helper art_of_def Let_def get_node[OF x1_in_Ai])
  from this \<gamma>(6) have sc:"?d \<Turnstile> ?pre \<and>\<^sub>f transition.formula \<tau>" by auto
  have formula_post:"formula ?post" using art[unfolded pre_art_body.art_def] b(1) by auto
  have formula_pre:"formula (?pre \<and>\<^sub>f transition.formula \<tau>)"
    using art[unfolded pre_art_body.art_def] l_isnode \<tau>_formula by auto
  have "transition_entailment (node_invariant (art_of Pi Ai) (state.location x1)) \<tau> 
                              (node_invariant (art_of Pi Ai) b)"
    using sim(2) l_isnode v b(2) unfolding pre_art_body.children_edges_cond_def by auto
  hence "?d \<Turnstile> ?post" using formula_pre formula_post assg_d sc
    unfolding transition_entailment_def implies_def by blast
  hence sat:"valuation y \<Turnstile> invariant (get_node Ai b)" using get_node[OF node_exists]
    unfolding art_of_def Let_def by(auto simp:map_of_helper)
  have "?trancl x' ?y" "?x_is_x ?y y" "?sat ?y"
    using y_is_y node_exists sat \<gamma>(2) path_is_trancl[OF path] by auto
  thus ?thesis by blast
qed

abbreviation sim_rel where
"sim_rel Ai x' x \<equiv> relabel_state ((art_node_impl.location \<circ>\<circ> get_node) Ai) x' = x \<and> satisfies_invariant Ai x'"

lemma path_simulation_simp:
  assumes wf:"pre_art_body.well_formed T (lts_of Pi) (art_of Pi Ai)"
      and lts:"lts (lts_of Pi)"
  shows "(x,y) \<in> transition (lts_of Pi) \<and> sim_rel Ai x' x \<Longrightarrow>
         \<exists> y'. (x',y') \<in> (transition (lts_of (lts_impl_of_art Pi Ai)))^+ \<and> sim_rel Ai y' y"
proof -
  assume ass:"(x,y) \<in> transition (lts_of Pi) \<and> sim_rel Ai x' x"
    hence tr:"(x,y) \<in> transition (lts_of Pi)"
      and sr: "relabel_state ((art_node_impl.location \<circ>\<circ> get_node) Ai) x' = x"
              "satisfies_invariant Ai x'" by auto
  from wf have sim:"pre_art_body.simulation_cond (lts_of Pi) (art_of Pi Ai)" and
               chld:"pre_art_body.children_edges_cond T (art_of Pi Ai)" and
               cov:"pre_art_body.cover_edges_cond T (art_of Pi Ai)" and
               art:"pre_art_body.art T (art_of Pi Ai)" and
               ini:"pre_art_body.initial_cond (art_of Pi Ai)"
          unfolding pre_art_body.well_formed_def by auto
  from path_simulation[OF tr sim chld cov art lts sr]
  show "\<exists> y'. (x',y') \<in> (transition (lts_of (lts_impl_of_art Pi Ai)))^+ \<and> sim_rel Ai y' y"
   by auto
qed

fun f_mimic where
  "f_mimic f s R x' 0 = x'" |
  "f_mimic f s R x' (Suc n) = (SOME y'. (f_mimic f s R x' n,y') \<in> s \<and> R y' (f (Suc n)))"

lemma f_mimic:
  assumes "\<And> x y x'. (x,y) \<in> r \<and> R x' x \<Longrightarrow> \<exists> y'. (x',y') \<in> s \<and> R y' y"
          "R x' (f 0)"
          "\<forall>i. (f i, f (Suc i)) \<in> r"
  shows "R (f_mimic f s R x' i) (f i) \<and> (f_mimic f s R x' i, f_mimic f s R x' (Suc i)) \<in> s"
using assms proof(induct i arbitrary: f x')
  case (Suc n f x')
    let "?c y'" = "(f_mimic f s R x' n, y') \<in> s \<and> R y' (f (Suc n))"
    let ?sy = "(SOME y'. ?c y')"
    have c1:"(f n, f (Suc n)) \<in> r" using Suc by auto
    have c2:"R (f_mimic f s R x' n) (f n)" using Suc by auto
    have "\<exists> y'. (f_mimic f s R x' n, y') \<in> s \<and> R y' (f (Suc n))" using c1 c2  Suc(2) by simp
    from someI_ex[OF this]
    have r:"R (f_mimic f s R x' (Suc n)) (f (Suc n))" by auto
    have "\<exists> y'. ?c y'" using assms(1) c1 c2 by auto
    from someI_ex[OF this] have c2:"R ?sy (f (Suc n))" by auto
    have c1:"(f (Suc n), f (Suc (Suc n))) \<in> r" using Suc by auto
    have "\<exists> y'. (?sy,y') \<in> s \<and> R y' (f (Suc (Suc n)))"
      using c1 c2 assms(1) by blast
    from someI_ex[OF this] r
    show ?case by auto
qed (auto simp:some_eq_ex[symmetric])

lemma simulation_SN_on:
  assumes "\<And> x y x'. (x,y) \<in> r \<and> R x' x \<Longrightarrow> \<exists> y'. (x',y') \<in> s \<and> R y' y"
          "\<forall> x \<in> X. \<exists> x' \<in> X'. R x' x"
          "SN_on s X'"
  shows "SN_on r X"
using assms(3) unfolding SN_on_def proof(rule contrapos_nn)
  assume "\<exists>f. f 0 \<in> X \<and> (\<forall>i. (f i, f (Suc i)) \<in> r)" then obtain f where
    f:"f 0 \<in> X" "(\<forall>i. (f i, f (Suc i)) \<in> r)" by auto
  from assms(2) f(1) obtain x' where x:"x' \<in> X'" "R x' (f 0)" by auto
  from f_mimic[of r R s x' f,OF assms(1) x(2) f(2), of "\<lambda> x y z. x"] f(1) x(1)
  have "f_mimic f s R x' 0 \<in> X' \<and> (\<forall>i. (f_mimic f s R x' i, f_mimic f s R x' (Suc i)) \<in> s)" by auto
  thus "\<exists>f. f 0 \<in> X' \<and> (\<forall>i. (f i, f (Suc i)) \<in> s)" by blast
qed

lemma stuttering_simulation_SN_on:
  assumes "\<And> x y x'. (x,y) \<in> r \<and> R x' x \<Longrightarrow> \<exists> y'. (x',y') \<in> s\<^sup>+ \<and> R y' y"
          "\<forall> x \<in> X. \<exists> x' \<in> X'. R x' x" "SN_on s X'"
  shows "SN_on r X"
by(rule simulation_SN_on[OF assms(1,2) SN_on_trancl[OF assms(3)]])

lemma state_for_state_2:
  fixes Pi::"('f, 'v, 't, 'l, 'tr) lts_impl"
  assumes wf:"pre_art_body.well_formed (T::('f,'t,'d) theory) (lts_of Pi) (art_of Pi Ai)"
   and   lts:"lts (lts_of Pi)"
   and   x:"x \<in> initial_states (lts_of Pi)"
   shows "\<exists>x'\<in>initial_states (lts_of (lts_impl_of_art Pi Ai)).
            State (valuation x') (art_node_impl.location (get_node Ai (state.location x'))) = x
            \<and> satisfies_invariant Ai x'"
proof(standard,standard)
  from wf have sim:"pre_art_body.simulation_cond (lts_of Pi) (art_of Pi Ai)" and
               ini:"pre_art_body.initial_cond (art_of Pi Ai)"
          unfolding pre_art_body.well_formed_def by auto
  from x have ass:"assignment (valuation x)" "State (valuation x) (lts_impl.initial Pi) = x"
     unfolding initial_states_def by auto
  let ?x' = "State (valuation x) (initial_node Ai)"
  show "?x' \<in> initial_states (lts_of (lts_impl_of_art Pi Ai))"
     unfolding lts_impl_of_art_def by(auto simp:initial_states_def ass)

  have "art.initial_node (art_of Pi Ai) \<in> nodes (art_of Pi Ai)"
    using ini[unfolded pre_art_body.initial_cond_def] by auto
  hence ini_ex:"node_exists Ai (art.initial_node (art_of Pi Ai))"
    by(auto simp:art_nodes_def art_of_def Let_def)
  have "node_location (art_of Pi Ai) (art.initial_node (art_of Pi Ai)) = lts.initial (lts_of Pi)"
    using sim[unfolded pre_art_body.simulation_cond_def] by auto
  hence h:"art_node_impl.location (get_node Ai (art_impl.initial_node Ai)) = lts_impl.initial Pi"
    using get_node(3)[OF ini_ex] by(auto simp:map_of_helper art_of_def Let_def)
  show "State (valuation ?x') (art_node_impl.location (get_node Ai (state.location ?x'))) = x"
    by(auto simp:h ass)
  have ini_ex3:"node_exists Ai (art_impl.initial_node Ai)"
    using ini[unfolded pre_art_body.initial_cond_def]
    by(auto simp:art_nodes_def art_of_def Let_def)
  have "node_invariant (art_of Pi Ai) (art.initial_node (art_of Pi Ai)) = form_True"
    using ini[unfolded pre_art_body.initial_cond_def] by auto
  hence "invariant (get_node Ai (art_impl.initial_node Ai)) = form_True"
    by (auto simp:art_of_def Let_def map_of_helper get_node(3)[OF ini_ex3])
  thus "satisfies_invariant Ai ?x'" by(auto simp:ass ini_ex3)
qed

lemma lts_termination_of_art:
  fixes Pi::"('f, 'v, 't, 'l, 'tr) lts_impl"
  assumes wf:"pre_art_body.well_formed (T::('f,'t,'d) theory) (lts_of Pi) (art_of Pi Ai)"
   and   lts:"lts (lts_of Pi)"
   and termn:"lts_termination (lts_of (lts_impl_of_art Pi Ai))"
  shows "lts_termination (lts_of Pi)"
by (rule stuttering_simulation_SN_on[OF path_simulation_simp[OF wf lts] _ termn]
   ;insert state_for_state_2[OF wf lts]; auto)

end

end