theory Safety_Checker
imports 
  Invariant_Checker
begin

definition (in pretheory) safe :: "('f,'v,'t,'l) lts \<Rightarrow> 'l \<Rightarrow> bool" where
  "safe P err \<equiv> \<forall> \<alpha>. State \<alpha> err \<notin> reachable_states P"

context pre_art_body
begin

definition "error_cond err \<equiv>
   \<forall> a \<in> nodes A. node_location A a = err \<longrightarrow> node_invariant A a \<longrightarrow>\<^sub>f form_False"

lemma safety_sound: 
  assumes well_formed
  and "lts P"
  and ec: "error_cond err"
  shows "safe P err" unfolding safe_def
proof (intro allI, rule)
  fix \<alpha>
  assume reach: "State \<alpha> err \<in> reachable_states P"
  from reachable_state[OF this] have \<alpha>: "assignment \<alpha>" by simp
  note [intro] = formula_node_invariant[OF assms(1)]
  thm disj_invariantD  well_formed_imp_disj_invariant[OF assms(1-2)]
  from disj_invariant_setD[OF well_formed_imp_disj_invariant[OF assms(1-2)] reach]
  obtain a where a: "a \<in> art.nodes A" "node_location A a = err" 
    and sat: "\<alpha> \<Turnstile> node_invariant A a" by auto
  from ec[unfolded error_cond_def, rule_format, OF a] \<alpha> sat a
  have "\<alpha> \<Turnstile> form_False" by auto
  then show False by auto
qed
end

context art_checker_body
begin

lemma error_cond_checker[iff]: "isOK(error_cond_checker err hints) \<Longrightarrow> error_cond err"
  unfolding error_cond_def error_cond_checker_def
  by (auto simp: ans zip_Cons split: option.splits dest: tc.formula_checker)

lemma art_safety_checker_main:
  "lts P \<Longrightarrow> isOK(art_safety_checker_main err h) \<Longrightarrow> safe P err"
  by (intro safety_sound, auto simp: art_safety_checker_main_def art_invariant_checker_main_def well_formed_def)

end


definition (in pre_art_checker) art_safety_checker ::
  "('f, 'v, 't, 'l::show, 'tr::show) lts_impl \<Rightarrow> ('f,'v,'t,'l,'n::show,'tr,'h list) art_impl \<Rightarrow> 'l \<Rightarrow> shows check" where
  "art_safety_checker Pi Ai err \<equiv>
   let P = lts_of Pi;
       A = art_of Pi Ai;
       H = make_hinter Pi Ai;
       eH = error_hints Ai
   in
   pre_art_checker_body.art_safety_checker_main T tc tc2 P A H err eH <+?
   (\<lambda> s. shows ''could not ensure safety w.r.t. error location '' o shows err o shows '' via ART:\<newline>'' o s)"

lemma (in art_checker) art_safety_checker:
  assumes Pi: "lts_impl Pi" and ok: "isOK (art_safety_checker Pi Ai err)"
  shows "safe (lts_of Pi) err" (is "safe ?P _")
proof -
  interpret art_checker_body T tc tc2 "?P" "art_of Pi Ai" "make_hinter Pi Ai"
    by (standard, auto simp: make_hinter_def)
  show ?thesis
    apply(rule art_safety_checker_main)
    using Pi ok[unfolded art_safety_checker_def]
    by (fast, auto)
qed

end