theory Termination_Checkers
  imports Linear_Integer_Inequalities Termination_Via_Invariants Safety_Checker
begin

text {* Here we instantiate "theory" by LIA. *}

locale LIA_checker begin

sublocale art_checker LIA entailment_checker entailment_checker
  unfolding art_checker_def by (auto simp: lin_ineq_theory.theory_checker)

type_synonym ('v, 'l) LIA_Inv = "(lin_ineq_op,'v,lin_ineq_type,'l) disjInv"
type_synonym ('v, 'l) LIA_Trans = "(lin_ineq_op,'v,lin_ineq_type,'l) transition"
type_synonym 'v trans_formula = "'v trans_var le_formula"

fun invariant_condition ::
  "('l \<Rightarrow> 'v le_exp)
  \<Rightarrow> ('tr::show \<Rightarrow> 'v::{compare,show} trans_formula
         \<Rightarrow> shows + nat list list \<times> nat list list)
  \<Rightarrow> ('l \<Rightarrow> 'v le_formula list)
  \<Rightarrow> 'tr \<times> ('v,'l) LIA_Trans
  \<Rightarrow> shows check"  where
"invariant_condition rf hint invar (tr,Transition src tgt \<phi>)
  = pre_theory_checker.check_transition entailment_checker (hint tr) (map (rename_vars Pre) (invar src)) \<phi>
                     (Fun Lesseq [rename_vars Post (rf tgt),rename_vars Pre (rf src)])
     <+? (\<lambda> s. shows ''Could not ensure a non-decrease of the transition '' \<circ> shows tr \<circ> shows '':\<newline>''\<circ> s)"

fun strict_condition ::
  "('l sharp \<Rightarrow> 'v le_exp)
  \<Rightarrow> ('tr::show \<Rightarrow> 'v::{compare,show} trans_formula
         \<Rightarrow> shows + nat list list \<times> nat list list)
  \<Rightarrow> ('l sharp \<Rightarrow> 'v le_formula list)
  \<Rightarrow> 'tr \<times> ('v,'l sharp) LIA_Trans
  \<Rightarrow> shows check"  where
"strict_condition rf hint invar (tr,Transition src tgt \<phi>)
  = pre_theory_checker.check_transition entailment_checker  (hint tr) (map (rename_vars Pre) (invar src)) \<phi>
                     (Fun Less [rename_vars Post (rf tgt),rename_vars Pre (rf src)])
     <+? (\<lambda> s. shows ''Could not ensure a strict-decrease of the transition '' \<circ> shows tr \<circ> shows '':\<newline>''\<circ> s)"

fun bound_condition ::
  "('l sharp \<Rightarrow> 'v le_exp)
  \<Rightarrow> int
  \<Rightarrow> ('tr::show \<Rightarrow> 'v::{compare,show} trans_formula
         \<Rightarrow> shows + nat list list \<times> nat list list)
  \<Rightarrow> ('l sharp \<Rightarrow> 'v le_formula list)
  \<Rightarrow> 'tr \<times> ('v,'l sharp) LIA_Trans
  \<Rightarrow> shows check"  where
"bound_condition rf bnd hint invar (tr,Transition src tgt \<phi>)
  = pre_theory_checker.check_transition entailment_checker  (hint tr) (map (rename_vars Pre) (invar src)) \<phi>
                     (Fun Less [Fun (Const bnd) [],rename_vars Pre (rf src)])
     <+? (\<lambda> s. shows ''Could not ensure a non-decrease of the transition '' \<circ> shows tr \<circ> shows '':\<newline>''\<circ> s)"

lemma check_transition_OK_mapped:
  assumes "isOK (tc.check_transition hinter (map (rename_vars Pre) disj_inv) \<phi> e)"
          "formula \<phi>" "list_all formula disj_inv" "\<phi>' \<in> set disj_inv"
  and sat: "\<alpha> \<Turnstile> \<phi>'"
  and "assignment \<alpha> \<and> assignment \<beta> \<and> assignment \<gamma>" "\<delta> \<alpha> \<beta> \<gamma> \<Turnstile> \<phi>" "is_bool e"
  shows "eval_true (\<delta> \<alpha> \<beta> \<gamma>) e"
proof -
  have "\<delta> \<alpha> \<beta> \<gamma> \<Turnstile> rename_vars Pre \<phi>'" using sat by auto
  from tc.check_transition_OK[OF assms(1) _ _ _ _ _ this assms(8)]
  show ?thesis using assms by auto
qed

lemma invariant_condition_isOK:
  assumes rf_ok:"\<And> v. type_of (rf v) = Some IntegerT"
  and is_ok:"isOK (invariant_condition rf hint I (tr, b))"
  and is_trans:"(State \<alpha> l, State \<beta> k) \<in> transition_step (snd (tr, b))"
  and *:"formula (transition.formula b)" "list_all formula (I l)" "\<phi> \<in> set (I l)" "\<alpha> \<Turnstile> \<phi>"
  shows "the_int (eval \<beta> (rf k)) \<le> the_int (eval \<alpha> (rf l))"
proof -
  let ?expr = "(Fun Lesseq [rename_vars Post (rf k),rename_vars Pre (rf l)])"
  have isOk:"source b = l \<Longrightarrow> target b = k \<Longrightarrow> isOK (pre_theory_checker.check_transition entailment_checker (hint tr) (map (rename_vars Pre) (I l)) (transition.formula b) ?expr)"
    using is_ok by (cases b;auto)
  from check_transition_OK_mapped[OF isOk *] is_trans[unfolded transition_step_def,simplified]
    show ?thesis by (auto simp:is_bool_def rf_ok eval_true_def)
qed

lemma strict_condition_isOK:
  assumes rf_ok:"\<And> v. type_of (rf v) = Some IntegerT"
  and is_ok:"isOK (strict_condition rf hint I (tr, b))"
  and is_trans:"(State \<alpha> l, State \<beta> k) \<in> transition_step (snd (tr, b))"
  and *:"formula (transition.formula b)" "list_all formula (I l)" "\<phi> \<in> set (I l)" "\<alpha> \<Turnstile> \<phi>"
  shows "the_int (eval \<beta> (rf k)) < the_int (eval \<alpha> (rf l))"
proof -
  let ?expr = "(Fun Less [rename_vars Post (rf k),rename_vars Pre (rf l)])"
  have isOk:"source b = l \<Longrightarrow> target b = k \<Longrightarrow> isOK (pre_theory_checker.check_transition entailment_checker (hint tr) (map (rename_vars Pre) (I l)) (transition.formula b) ?expr)"
    using is_ok by (cases b;auto)
  from check_transition_OK_mapped[OF isOk *] is_trans[unfolded transition_step_def,simplified]
    show ?thesis by (auto simp:is_bool_def rf_ok eval_true_def)
qed

lemma bound_condition_isOK:
  assumes rf_ok:"\<And> v. type_of (rf v) = Some IntegerT"
  and is_ok:"isOK (bound_condition rf bnd hint I (tr, b))"
  and is_trans:"(State \<alpha> l, State \<beta> k) \<in> transition_step (snd (tr, b))"
  and *:"formula (transition.formula b)" "list_all formula (I l)" "\<phi> \<in> set (I l)" "\<alpha> \<Turnstile> \<phi>"
  shows "bnd < the_int (eval \<alpha> (rf l))"
proof -
  let ?expr = "(Fun Less [Fun (Const bnd) [],rename_vars Pre (rf l)])"
  have isOk:"source b = l \<Longrightarrow> target b = k \<Longrightarrow> isOK (pre_theory_checker.check_transition entailment_checker (hint tr) (map (rename_vars Pre) (I l)) (transition.formula b) ?expr)"
    using is_ok by (cases b;auto)
  from check_transition_OK_mapped[OF isOk *] is_trans[unfolded transition_step_def,simplified]
    show ?thesis by (auto simp:is_bool_def rf_ok eval_true_def)
qed

end

interpretation non_inf_int: non_inf_order "{(x::int, y). x > y}" "{(x, y). x \<ge> y}"
  by(unfold_locales,insert non_inf_int_gt,auto simp:trans_def refl_on_def)

locale int_rf = LIA_checker
  for rf :: "('l::{show}) sharp \<Rightarrow> (lin_ineq_op, 'b :: {compare,linorder,show} \<times> lin_ineq_type) Term.term"
  and bound :: "int" +
  assumes rf_ok:"\<And> v. type_of (rf v) = Some IntegerT"
begin
  interpretation LIA_edge_checker : delete_edge_checker LIA entailment_checker entailment_checker rf
     the_int "{(x,y). y < x}" "{(x,y). y \<le> x}" bound "invariant_condition rf hint" "strict_condition rf hint" "bound_condition rf bound hint"
  proof
    let ?S = "{(x,y). y < x}::(int\<times>int) set" let ?NS = "{(x,y). y \<le> x}::(int\<times>int) set"
    let ?S2 = "{(x, y). 0 \<le> x \<and> y < x}::(int\<times>int) set"
    from int_SN.SN have s2:"SN ?S2" by fast
    { fix tr \<tau> I s t \<phi>
      let ?s = "the_int (eval (valuation s) (rf (state.location s)))"
      let ?t = "the_int (eval (valuation t) (rf (state.location t)))"
      assume *:
             "(s, t) \<in> transition_step \<tau>"
             "local.formula (transition.formula \<tau>)"
             "\<phi> \<in> set (I (state.location s))"
             "is_sharp_transition \<tau>"
             "valuation s \<Turnstile> \<phi>"
             "\<forall>l. list_all local.formula (I l)"
      { assume o2:"isOK (invariant_condition rf hint I (tr, \<tau>))"
        show "(?s, ?t) \<in> ?NS \<union> ?S"
          using invariant_condition_isOK[OF rf_ok o2 _ *(2) _ *(3)] * by auto }
      { assume o2:"isOK (strict_condition rf hint I (tr, \<tau>))"
        show "(?s, ?t) \<in> ?S"
          using strict_condition_isOK[OF rf_ok o2 _ *(2) _ *(3)] * by auto}
      { assume o2:"isOK (bound_condition rf bound hint I (tr, \<tau>))"
        show "(?s, bound) \<in> ?S"
          using bound_condition_isOK[OF rf_ok o2 _ *(2) _ *(3)] * by (cases t;auto)}
    }
  qed
  
  abbreviation invar_reduction where
  "invar_reduction hint lst Ai \<equiv>
    no_invariants (with_invariants_reduction (LIA_edge_checker.reduction hint lst)
                                             (no_modification_create (art_invariant_checker Ai)))"

  (* procedure for making structures like this:
     - start by typing "thm disj_invariant_checker" or something similar
     - add structure by using lemma_name[OF ...]
     - copy the resulting condition.
       At this point, you should be able to prove a lemma such as the one below
     - create an abbreviation or definition to make the theorem look beautiful,
       repair the proof if necessary
  *)
  lemma lts_termination:
  assumes "isOK (cooperation_cond [invar_reduction hint lst Ai] P R)"
  shows "lts_termination (lts_of R)"
  using checker_is_ok[OF assms]
        no_invariants_is_cooperation_SN_preserving[OF
           with_invariants_reduction_is_create[OF
              no_modification_is_disj_invariant_create[OF disj_invariant_checker]
              LIA_edge_checker.all_in_reduction_disj_invariant_update]]
     by auto

end


end