(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2009-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2009-2015)
Author:  Akihisa Yamada <akihisa.yamada@uibk.ac.at> (2016)
License: LGPL (see file COPYING.LESSER)
*)
theory Relative_Usable_Rules
imports
  "../Rewriting/Trs_Impl"
  "../Orderings/Reduction_Pair"
  "../AC_Rewriting/AC_Rewriting"
begin

subsection {* Trivia *}
lemma Fun_nrrsteps:
  assumes st: "(Fun f ss,t) \<in> (nrrstep R)\<^sup>*"
  shows "\<exists>ts. t = Fun f ts \<and> length ts = length ss \<and> (\<forall>i<length ss. (ss ! i, ts ! i) \<in> (rstep R)\<^sup>*)"
proof-
  from nrrsteps_preserve_root[OF st] obtain ts where [simp]: "t = Fun f ts" by auto
  from nrrsteps_equiv_num_args[OF st] nrrsteps_imp_arg_rsteps[OF st]
  show ?thesis by auto
qed

lemma nrrstep_subst:
  assumes "(s,t) \<in> nrrstep R"
  shows "(s\<cdot>\<sigma>, t\<cdot>\<sigma>) \<in> nrrstep R"
proof-
  from assms obtain C \<tau> l r
  where lr: "(l,r) \<in> R" and s: "s = C\<langle>l\<cdot>\<tau>\<rangle>" and t: "t = C\<langle>r\<cdot>\<tau>\<rangle>" and C: "C \<noteq> \<box>" by (elim nrrstepE)
  moreover from s have "s\<cdot>\<sigma> = (C\<cdot>\<^sub>c\<sigma>)\<langle>l \<cdot> \<tau> \<circ>\<^sub>s \<sigma>\<rangle>" by auto
  moreover from t have "t\<cdot>\<sigma> = (C\<cdot>\<^sub>c\<sigma>)\<langle>r \<cdot> \<tau> \<circ>\<^sub>s \<sigma>\<rangle>" by auto
  moreover from C have "C\<cdot>\<^sub>c\<sigma> \<noteq> \<box>" using subst_apply_ctxt.elims by auto
  ultimately show ?thesis by (intro nrrstepI)
qed


lemma SN_on_relto_Image_S:
  assumes "SN_on (relto R S) X"
  shows "SN_on (relto R S) (S `` X)"
proof
  fix seq assume 0: "seq 0 \<in> S `` X" and 1: "chain (relto R S) seq"
  from 0 obtain x where 2: "x \<in> X" and 3: "(x,seq 0) \<in> S" by auto
  let ?seq = "\<lambda>i. if i = 0 then x else seq i"
  from 3 1 have "(?seq 0, ?seq 1) \<in> S O relto R S" by auto
  then have "(?seq 0, ?seq 1) \<in> relto R S" by regexp
  with 1 have "chain (relto R S) ?seq" by auto
  moreover from 2 have "?seq 0 \<in> X" by auto
  ultimately show False using assms by fast
qed

lemma SN_on_relto_Image_R:
  assumes "SN_on (relto R S) X"
  shows "SN_on (relto R S) (R `` X)"
proof
  fix seq assume 0: "seq 0 \<in> R `` X" and 1: "chain (relto R S) seq"
  from 0 obtain x where 2: "x \<in> X" and 3: "(x,seq 0) \<in> R" by auto
  let ?seq = "\<lambda>i. case i of 0 \<Rightarrow> x | Suc i \<Rightarrow> seq i"
  from 3 have "(?seq 0, ?seq 1) \<in> relto R S" by auto
  with 1 have "(?seq i, ?seq (Suc i)) \<in> relto R S" for i by(cases i, auto)
  then have "chain (relto R S) ?seq" by auto
  moreover from 2 have "?seq 0 \<in> X" by auto
  ultimately show False using assms by fast
qed

lemma image_rtrancl_expand: "{y. (x,y) \<in> R\<^sup>*} = {x} \<union> (\<Union> y \<in> R `` {x}. {z. (y,z) \<in> R\<^sup>*})" (is "?L = ?R")
proof(intro equalityI subsetI)
  fix z
  assume "z \<in> ?L"
  then have "(x,z) \<in> R\<^sup>*" by auto
  then obtain n where xz: "(x,z) \<in> R^^n" by auto
  show "z \<in> ?R"
  proof (cases n)
    case 0 with xz show ?thesis by auto
  next
    case (Suc m)
      from xz[unfolded this] obtain y where xy: "(x,y) \<in> R" and yz: "(y,z) \<in> R^^m" by (elim relpow_Suc_E2)
      from yz relpow_imp_rtrancl have "z \<in> R\<^sup>* `` {y}" by auto
      with xy show ?thesis by auto
  qed
qed auto

lemma finite_ImageI:
  assumes R: "\<And>x. x \<in> X \<Longrightarrow> finite {y. (x,y) \<in> R}" and X: "finite X"
  shows "finite (R `` X)"
proof-
  have "R `` X = (\<Union>x \<in> X. {y. (x,y) \<in> R})" by auto
  also with R have "finite ..." unfolding finite_UN[OF X] by auto
  finally show ?thesis by auto
qed

lemma finite_rule_reducts:
  assumes "vars_term r \<subseteq> vars_term l"
  shows "finite {t. (s, t) \<in> rstep {(l, r)}}" (is "finite ?T")
proof -
  def h \<equiv> "\<lambda>p. (ctxt_of_pos_term p s)\<langle>r \<cdot> the (match (s |_ p) l)\<rangle>" -- \<open>computes reduct from position\<close>
  def P \<equiv> "{p \<in> poss s. \<exists>\<sigma>. match (s |_ p) l = Some \<sigma>}" -- \<open>positions of subterms matching \<open>l\<close>\<close>
  { fix C \<sigma> assume [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    then obtain \<tau> where *: "match (s |_ hole_pos C) l = Some \<tau>"
      and [simp]: "r \<cdot> \<sigma> = r \<cdot> \<tau>"
      using assms and match_complete' by (force simp: term_subst_eq_conv)
    then have "h (hole_pos C) = C\<langle>r \<cdot> \<sigma>\<rangle>" by (auto simp: h_def)
    moreover have "h (hole_pos C) \<in> h ` P" using * by (auto simp: P_def)
    ultimately have "C\<langle>r \<cdot> \<sigma>\<rangle> \<in> h ` P" by simp }
  then have "?T = h ` P" by (auto intro!: rstepI simp: h_def P_def match_matches ctxt_supt_id)
  moreover have "finite P" by (rule finite_subset [of _ "poss s"]) (auto simp: P_def)
  ultimately show ?thesis by simp
qed

lemma finite_rstep_reducts:
  assumes "finite R"
    and "\<forall>(l, r) \<in> R. vars_term r \<subseteq> vars_term l"
  shows "finite {t. (s, t) \<in> rstep R}" (is "finite ?T")
proof -
  have "?T = \<Union>((\<lambda>(l, r). {t. (s, t) \<in> rstep {(l, r)}}) ` R)" by auto
  then show ?thesis
    using assms by (auto simp: finite_rule_reducts)
qed

lemma finite_qrstep_reducts:
  assumes "finite (applicable_rules Q R)"
    and "wwf_qtrs Q R"
  shows "finite {t. (s,t) \<in> qrstep False Q R}" (is "finite ?T")
proof-
  have "?T = {t. (s,t) \<in> qrstep False Q (applicable_rules Q R)}" by (auto simp: qrstep_applicable_rules)
  also have "finite ..."
  proof(rule finite_subset)
    show "... \<subseteq> {t. (s, t) \<in> rstep (applicable_rules Q R)}" by auto
    show "finite ..." by(rule finite_rstep_reducts, insert assms, auto simp: wwf_qtrs_def applicable_rules_def)
  qed
  finally show ?thesis .
qed

lemma size_ctxt_compose_gt: assumes "D \<noteq> \<box>" shows "size (C \<circ>\<^sub>c D) > size C"
proof-
  have "size (C \<circ>\<^sub>c D) \<ge> size C \<and> (D = \<box> \<or> ?thesis)" by(induct C, cases D, auto)
  with assms show ?thesis by auto
qed

lemma size_ctxt_apply_ge: "size (C\<langle>s\<rangle>) \<ge> size C + size s" (* not equal! *) by(induct C, auto)

subsection {* Quasi-termination *}

definition "quasi_terminating_on R X \<equiv> \<forall>x \<in> X. finite {y. (x,y) \<in> R\<^sup>*}"

lemma quasi_terminating_onI[intro]:
  assumes "\<And>x. x \<in> X \<Longrightarrow> finite {y. (x,y) \<in> R\<^sup>*}"
  shows "quasi_terminating_on R X"
  using assms by (auto simp: quasi_terminating_on_def)

lemma quasi_terminating_onD[dest]:
  assumes "quasi_terminating_on R X" and "x \<in> X"
  shows "finite {y. (x,y) \<in> R\<^sup>*}"
  using assms[unfolded quasi_terminating_on_def] by auto

abbreviation "quasi_terminating R \<equiv> quasi_terminating_on R UNIV"

text {* Quasi-terminating TRSs *}

locale quasi_terminating_trs =
  fixes S
  assumes QT: "quasi_terminating (rstep S)"
begin

fun accumulate_ctxts where
  "accumulate_ctxts C 0 = \<box>"
  | "accumulate_ctxts C (Suc i) = accumulate_ctxts C i \<circ>\<^sub>c C i"

lemma SN_suptrel:
  "SN (suptrel S)"
proof(unfold suptrel_def, rule SN_on_trancl, rule SN_onI)
  let ?R = "(rstep S)\<^sup>*"
  fix seq assume chain: "chain (?R O {\<rhd>} O ?R) seq"
  have "\<forall>i. \<exists>C. C \<noteq> \<box> \<and> (seq i, C\<langle>seq (Suc i)\<rangle>) \<in> ?R" (is "\<forall>i. ?goal i")
  proof
    fix i
    from chain have "(seq i, seq (Suc i)) \<in> (?R O {\<rhd>} O ?R)" by auto
    then obtain s t
    where s: "(seq i, s) \<in> ?R"
      and st: "(s,t) \<in> {\<rhd>}"
      and t: "(t, seq (Suc i)) \<in> ?R" by auto
    from st obtain C where "s = C\<langle>t\<rangle>" and C: "C\<noteq>\<box>" by auto
    moreover with s have "(seq i,C\<langle>t\<rangle>) \<in> ?R" by auto
    moreover with t rsteps_closed_ctxt have "(C\<langle>t\<rangle>,C\<langle>seq (Suc i)\<rangle>) \<in> ?R" by auto
    ultimately show "?goal i" by auto
  qed
  from choice[OF this] obtain C where *: "\<And>i. C i \<noteq> \<box> \<and> (seq i, (C i)\<langle>seq (Suc i)\<rangle>) \<in> ?R" by auto
  let ?s = "\<lambda>i. (accumulate_ctxts C i)\<langle>seq i\<rangle>"
  def m \<equiv> "Max { size t | t. (?s 0, t) \<in> ?R }"
  moreover
    have "(?s 0, ?s i) \<in> ?R" for i
    proof (induct i)
      case (Suc i)
      moreover from * have "(?s i, ?s (Suc i)) \<in> ?R" by (auto intro: rsteps_closed_ctxt)
      ultimately show ?case by auto
    qed auto
    note this[of "Suc m"]
  moreover
    from QT have "finite { t. (?s 0, t) \<in> ?R }" by auto
    then have "finite { size t | t. (?s 0, t) \<in> ?R }" by auto
  ultimately have "size (?s (Suc m)) \<le> m" by(auto intro: Max_ge)
  moreover
    have "size (?s i) \<ge> i" for i
    proof-
      have "size (accumulate_ctxts C i) \<ge> i"
      proof (induct i)
        case (Suc i)
        from *[of i] have "size (accumulate_ctxts C (Suc i)) > size (accumulate_ctxts C i)"
          by(auto intro: size_ctxt_compose_gt)
        with Suc show ?case by auto
      qed auto
      with size_ctxt_apply_ge[of "accumulate_ctxts C i" "seq i"]
      show ?thesis by auto
    qed
    note this[of "Suc m"]
  ultimately show False by auto
qed

lemma SN_supt_relto: "SN (relto {\<rhd>} (rstep S))"
  by (rule SN_subset[OF SN_suptrel], unfold suptrel_def, auto)

lemma SN_supt: "SN ((rstep S)\<^sup>* O {\<rhd>})"
  by (rule SN_subset[OF SN_supt_relto], auto)

lemma QT_relstep:
  assumes SN: "SN_on (relstep R S) X"
      and fin: "finite R"
      and wf: "\<forall> (l,r) \<in> R. vars_term r \<subseteq> vars_term l"
    shows "quasi_terminating_on (relstep R S) X"
proof
  fix x assume "x \<in> X"
  with SN
  show "finite {y. (x,y) \<in> (relstep R S)\<^sup>*}"
  proof(induct x rule: SN_on_induct)
    case (IH x)
      have "{y. (x,y) \<in> (relstep R S)\<^sup>*} = {x} \<union> (\<Union> y \<in> relstep R S `` {x}. {z. (y,z) \<in> (relstep R S)\<^sup>*})"
        by(rule image_rtrancl_expand)
      also have "finite ..."
      proof (intro finite_UnI conjI finite_UN_I)
        from IH show "\<And>y. y \<in> relstep R S `` {x} \<Longrightarrow> finite {z. (y,z) \<in> (relstep R S)\<^sup>*}" by (auto simp: Image_def)
        have "finite ((rstep S)\<^sup>* `` rstep R `` (rstep S)\<^sup>* `` {x})"
        proof (rule finite_ImageI)
          show "finite (rstep R `` (rstep S)\<^sup>* `` {x})"
          proof (rule finite_ImageI)
            from QT show "finite ((rstep S)\<^sup>* `` {x})" by (auto simp: Image_def)
          qed (insert finite_rstep_reducts[OF fin wf], auto)
        qed (insert QT, auto simp: Image_def)
        then show "finite (relstep R S `` {x})" by (auto simp: relcomp_Image)
      qed auto
      finally show ?case.
  qed
qed

end

subsection {* Connectability *}

text {* The following notion characterizes what one estimates by TCAP-unifiability etc. *}

definition "connectable R s t \<equiv> \<exists>\<sigma> \<tau>. (s\<cdot>\<sigma>, t\<cdot>\<tau>) \<in> (nrrstep R)\<^sup>*"

lemma connectableI[intro]: fixes \<sigma> \<tau>
  shows "(s\<cdot>\<sigma>, t\<cdot>\<tau>) \<in> (nrrstep R)\<^sup>* \<Longrightarrow> connectable R s t" unfolding connectable_def by auto

lemma connectableE[elim]:
  assumes "connectable R s t" obtains \<sigma> \<tau> where "(s\<cdot>\<sigma>, t\<cdot>\<tau>) \<in> (nrrstep R)\<^sup>*"
  using assms[unfolded connectable_def] by auto

lemma connectable_stable:
  assumes con: "connectable R (s\<cdot>\<sigma>) t" shows "connectable R s t"
proof-
  from con obtain \<sigma>' \<tau> where "(s \<cdot> (\<sigma> \<circ>\<^sub>s \<sigma>'), t\<cdot>\<tau>) \<in> (nrrstep R)\<^sup>*" by auto
  from connectableI[OF this] show ?thesis by auto
qed

text{* Checks if a term is @{term connectable} to a usable rule. *}

definition i_trans_check where
  "i_trans_check R ur s \<equiv> is_Var s \<or> (\<forall> (l,r) \<in> R. connectable R s l \<longrightarrow> (l,r) \<in> ur)"

lemma i_trans_checkI[intro]:
  assumes "is_Fun s \<Longrightarrow> (\<And>l r. (l,r) \<in> R \<Longrightarrow> connectable R s l \<Longrightarrow> (l,r) \<in> ur)"
  shows "i_trans_check R ur s"
  unfolding i_trans_check_def using assms by auto

lemma i_trans_checkD[dest]:
  assumes "i_trans_check R ur s"
  shows "is_Fun s \<Longrightarrow> (l,r) \<in> R \<Longrightarrow> connectable R s l \<Longrightarrow> (l,r) \<in> ur"
  using assms unfolding i_trans_check_def by auto

lemma i_trans_check_var[simp]: "i_trans_check R ur (Var x)" by (auto simp: i_trans_check_def)
lemma i_trans_check_fun: "i_trans_check R ur (Fun f ss) \<longleftrightarrow> (\<forall> (l,r) \<in> R. connectable R (Fun f ss) l \<longrightarrow> (l,r) \<in> ur)"
  by (auto simp: i_trans_check_def)

lemma i_trans_check_stable:
  assumes "is_Fun s" "i_trans_check R ur s" shows "i_trans_check R ur (s \<cdot> \<sigma>)"
proof
  fix l r
  assume ss: "is_Fun (s \<cdot> \<sigma>)" and lr: "(l,r) \<in> R" and sl: "connectable R (s \<cdot> \<sigma>) l"
  from connectable_stable[OF sl] have "connectable R s l" by auto
  with assms lr show "(l,r) \<in> ur" by auto
qed


subsection {* Main *}

definition qrewrite :: "bool \<Rightarrow> ('f,'v)terms \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term list"
  where "qrewrite nfs Q R t \<equiv> filter (\<lambda> s. (t,s) \<in> qrstep nfs Q (set R)) (rewrite R t)" 

lemma qrewrite_sound: assumes "t \<in> set (qrewrite nfs Q R s)"
  shows "(s,t) \<in> qrstep nfs Q (set R)" using assms unfolding qrewrite_def by auto

lemma qrewrite_complete:
  assumes SN: "SN_on (qrstep nfs Q (set R)) {s}"
  and qstep: "(s,t) \<in> qrstep nfs Q (set R)"
  and nfs: "nfs \<Longrightarrow> Q \<noteq> {} \<Longrightarrow> wwf_qtrs Q (set R)"
  shows "t \<in> set (qrewrite nfs Q R s)" 
  using qstep
proof 
  fix C \<sigma> l r
  assume Q: "\<forall> u \<lhd> l \<cdot> \<sigma>. u \<in> NF_terms Q" and lr: "(l,r) \<in> set R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  from wwf_qtrs_imp_nfs_False_switch[OF nfs, of nfs] have switch:
    "qrstep nfs Q (set R) = qrstep False Q (set R)" .
  note SN = SN[unfolded switch]
  note qstep = qstep[unfolded switch]
  from only_applicable_rules[OF Q] SN_on_imp_wwf_rule[OF SN s lr Q, unfolded wwf_rule_def] have vars: "vars_term r \<subseteq> vars_term l" by auto
  have step: "(s,t) \<in> rstep_r_p_s (set R) (l,r) (hole_pos C) \<sigma>"
    unfolding rstep_r_p_s_def Let_def s t
    by (simp add: lr)
  have step: "t \<in> set (rewrite R s)"
    by (rule rewrite_complete[OF step vars])
  thus ?thesis using qstep unfolding qrewrite_def switch by auto
qed



definition qrel where "qrel Q R S \<equiv> relto (qrstep False Q R) (rstep S)"

lemma mem_qrelI: "p \<in> relto (qrstep False Q R) (rstep S) ==> p \<in> qrel Q R S" unfolding qrel_def by auto

lemma qrstep_subset_qrel: "qrstep False Q R \<subseteq> qrel Q R S" unfolding qrel_def by auto

lemma qrel_empty_is_qrstep: "qrel Q R {} = qrstep False Q R"
  unfolding qrel_def by auto

lemma ctxt_closed_qrel: "ctxt.closed (qrel Q R E)"
  unfolding qrel_def
  by (rule ctxt.closed_relto[OF ctxt_closed_qrstep ctxt_closed_rstep]) 

fun comb :: "('f \<times> nat) \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term" where
    "comb cn [] = Var undefined"
  | "comb (c,n) (t # ts) = Fun c (t # comb (c,n) ts # replicate n (Var undefined))"

lemma ce_trs_sound: "t \<in> set ts \<Longrightarrow> (comb cn ts, t) \<in> (rstep (ce_trs cn))^+"
using assms 
proof (induct ts, simp)
  case (Cons s ss)
  obtain c n where cn: "cn = (c,n)" by force
  show ?case 
  proof (cases "s = t")
    case True 
    with cn have "(comb cn (s # ss),s) \<in> ce_trs cn" by (simp add: ce_trs.simps)
    hence "(comb cn (s # ss),t) \<in> rstep (ce_trs cn)" using True by blast
    thus ?thesis using True by force
  next
    case False
    hence ind: "(comb cn ss, t) \<in> (rstep (ce_trs cn))^+" using Cons by auto
    from cn have "(comb cn (s # ss),comb cn ss) \<in> ce_trs cn" by (simp add: ce_trs.simps)
    hence "(comb cn (s # ss),comb cn ss) \<in> rstep (ce_trs cn)" by blast 
    with ind show ?thesis by auto
  qed
qed

lemma comb_base:
  shows "(comb cn ss,Var undefined) \<in> (rstep (ce_trs cn))\<^sup>*" (is "_ \<in> ?rel")
proof(cases cn,induct ss)
  case Nil then show ?case by auto
next
  case (Cons s ss)
  then have "(comb cn ss,Var undefined) \<in> ?rel" by auto
  moreover have "(comb cn (s#ss), comb cn ss) \<in> ?rel" by(auto simp: Cons(2) ce_trs.simps)
  ultimately show ?case by auto
qed

context linorder begin

definition "sorted_distinct_list X \<equiv> (SOME xs. set xs = X \<and> sorted xs \<and> distinct xs)"

lemma finite_set_as_list: "finite X \<Longrightarrow> \<exists>xs. set xs = X \<and> sorted xs \<and> distinct xs"
proof(induct rule:finite_induct)
  case empty show ?case by auto
next
  case (insert x X)
    then obtain xs where [simp]: "X = set xs" and "sorted xs" "distinct xs" by auto
    with insert(2) show ?case
      apply (intro exI[of _ "insort_insert x xs"] conjI)
      apply (simp add: set_insort_insert)
      using sorted_insort_insert apply force
      using distinct_insort_insert apply force
      done
qed

lemma sorted_distinct_list:
  assumes "finite X"
  shows "sorted (sorted_distinct_list X)" "distinct (sorted_distinct_list X)" "set (sorted_distinct_list X) = X"
  using someI_ex[OF finite_set_as_list[OF assms]] unfolding sorted_distinct_list_def by auto

end

locale itrans = 
  fixes Q :: "('f,'v)terms"
  and R :: "('f,'v)trs"
  and S :: "('f,'v)trs"
  and ur :: "('f,'v)trs"
  and cn :: "('f \<times> nat)"
  assumes QT: "quasi_terminating (rstep S)"
  and wf: "wwf_qtrs Q R"
  and fin: "finite (applicable_rules Q R)"
begin

interpretation quasi_terminating_trs S using QT unfolding quasi_terminating_trs_def.

definition LT :: "('f,'v)term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" where "LT = undefined"
definition "LE s t \<equiv> s = t \<or> LT s t"

sublocale linorder LE LT sorry

lemma ce_trs_subset:
  assumes "set ss \<supseteq> set ts"
      and "sorted ss" "distinct ss"
      and "sorted ts" "distinct ts"
  shows "(comb cn ss, comb cn ts) \<in> (rstep (ce_trs cn))\<^sup>*" (is "_ \<in> ?rel")
  using assms
proof(induct ss arbitrary: ts)
  case Nil then show ?case by auto
next
  case (Cons s ss)
  then show ?case
  proof(induct ts)
    case Nil then show ?case by (auto intro!: comb_base)
  next
    case IH: (Cons t ts)
      then have ss: "sorted ss" "distinct ss" and ts: "sorted ts" "distinct ts" using sorted_Cons by auto
      obtain c n where cn: "cn = (c,n)" by (cases cn)
      show ?case
      proof(cases "s = t")
        case True
        with IH have "set ss \<supseteq> set ts" by auto
        from IH(2)[OF this] ss ts have "(comb cn ss, comb cn ts) \<in> ?rel" by auto
        from rsteps_closed_ctxt[OF this, of "More c [t] \<box> (replicate n (Var undefined))"]
        show ?thesis unfolding True cn by auto
      next
        case False
        moreover with IH(3) have tss: "t \<in> set ss" by auto
        moreover from IH(4) sorted_Cons have "\<forall>s' \<in> set ss. LE s s'" by auto
        ultimately have "LT s t" by auto
        with IH(6)[unfolded sorted_Cons]
        have "\<forall>t' \<in> set ts. LT s t'" by auto
        with IH(3) have "set ss \<supseteq> set ts" by auto
        with tss have "set ss \<supseteq> set (t#ts)" by auto
        from IH(2)[OF this] ss IH(6,7) have "(comb cn ss, comb cn (t#ts)) \<in> ?rel" by auto
        moreover have "(comb cn (s#ss), comb cn ss) \<in> ?rel" by (auto simp:cn ce_trs.simps)
        ultimately show ?thesis by auto
      qed
    qed
qed

definition qr_successors :: "('f,'v)term \<Rightarrow> ('f,'v)term list" where
  "qr_successors s \<equiv> SOME list. set list = {t. (s,t) \<in> qrstep False Q R}"

lemma qr_successors[simp]: assumes "SN_on (qrstep False Q R) {t}"
  shows "set (qr_successors s) = {t. (s,t) \<in> qrstep False Q R}"
proof -
  from wf fin have "\<exists> xs. set xs = {t. (s,t) \<in> qrstep False Q R}"
    by(intro finite_list finite_qrstep_reducts, auto)
  from someI_ex[OF this,folded qr_successors_def] show ?thesis.
qed

function i_trans :: "('f,'v)term \<Rightarrow> ('f,'v)term" and i_trans_sub
where "i_trans (Var x) = (Var x)" 
   |  "i_trans (Fun f ss) = (
       if SN_on (qrel Q R S) {Fun f ss} then
         if i_trans_check (R \<union> S) ur (Fun f ss) then Fun f (map i_trans ss)
         else (comb cn) (sorted_distinct_list (i_trans_sub ` (rstep S)\<^sup>* `` {Fun f ss}))
       else Fun f ss)"
  | "i_trans_sub (Var x) = Var x"
  | "i_trans_sub (Fun f ss) = (
     if SN_on (qrel Q R S) {Fun f ss} then
       if i_trans_check (R \<union> S) ur (Fun f ss) then Fun f (map i_trans ss)
       else (comb cn) (Fun f (map i_trans ss) # map i_trans (qr_successors (Fun f ss)))
     else Fun f ss)"
by (pat_completeness, auto)


termination i_trans
proof-
  interpret E_compatible "qrel Q R S" "(rstep S)^*" 
    by (standard, auto simp: qrel_def)
  let ?M = "{ (Inl s, Inr t) | s t. (s,t) \<in> (rstep S)\<^sup>* } \<union>
            { (Inl s, Inl t) | s t. (s,t) \<in> restrict_SN_supt_E } \<union>
            { (Inr s, Inl t) | s t. (s,t) \<in> restrict_SN_supt_E }"
  have "SN (restrict_SN_supt_E)"
    by (rule ctxt_closed_imp_SN_restrict_SN_E_supt[OF ctxt_closed_qrel SN_supt_relto])
  then have "SN ?M" sorry
  hence wf: "wf (?M\<inverse>)" (is "wf ?R")  by (simp add: SN_iff_wf)
  have "\<And>X. SN_on (qrel Q R S) X \<Longrightarrow> SN_on (qrstep False Q R) X" by(rule SN_on_mono,unfold qrel_def,auto)
  note [simp] = qr_successors[OF this]
  show ?thesis by (standard, rule wf, auto intro!: restrict_SN_supt_E_I[OF ctxt_closed_qrel] intro: mem_qrelI)
qed


abbreviation
  i_trans_subst ::
    "('f, 'v) subst \<Rightarrow> ('f, 'v) subst"
where
  "i_trans_subst \<sigma> \<equiv> (\<lambda>x. i_trans (\<sigma> x))"

end

fun ur_closed_term_af :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "ur_closed_term_af _ _ _ (Var x) = True"
   |  "ur_closed_term_af R ur \<pi> (Fun f ts) =
   ( (\<forall> i < length ts. i \<in> \<pi> (f, (length ts)) \<longrightarrow> ur_closed_term_af R ur \<pi> (ts ! i)) \<and> 
     i_trans_check R ur (Fun f ts)
)"

abbreviation ur_closed_af :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> bool"
where "ur_closed_af R ur \<pi> \<equiv> \<forall> (l,r) \<in> ur. ur_closed_term_af R ur \<pi> r"

abbreviation ur_P_closed_af :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> ('f,'v)trs \<Rightarrow> bool" 
  where "ur_P_closed_af R ur \<pi> P  \<equiv> \<forall> st \<in> P. ur_closed_term_af R ur \<pi> (snd st)"  

lemma i_trans_check_nrrstep:
  assumes ur_closed: "ur_closed_af R ur \<pi>"
  and s: "i_trans_check R ur s"
  and st: "(s,t) \<in> nrrstep R"
  and sFun: "is_Fun s"
  shows "i_trans_check R ur t"
proof
  fix l r assume "is_Fun t" and lr: "(l,r) \<in> R" and connect: "connectable R t l"
  from connect obtain \<sigma> \<tau> where "(t\<cdot>\<sigma>,l\<cdot>\<tau>) \<in> (nrrstep R)\<^sup>*" by (elim connectableE)
  moreover from nrrstep_subst[OF st] have "(s\<cdot>\<sigma>,t\<cdot>\<sigma>) \<in> nrrstep R" by auto
  ultimately have "connectable R s l" by (intro connectableI[of _ \<sigma> _ \<tau>], auto)
  with i_trans_checkD[OF s sFun lr] show "(l,r)\<in> ur" by auto
qed

notation Ground_Context.equiv_class ("[_]")

lemma SN_on_stable:
  assumes stable: "\<And>x y. x \<in> R\<^sup>* `` X \<Longrightarrow> (x,y) \<in> R \<Longrightarrow> (f x, f y) \<in> R" and SN: "SN_on R (f ` X)" shows "SN_on R X"
proof(intro SN_onI)
  fix seq assume 0: "seq 0 \<in> X" and chain: "chain R seq"
  show False
  proof(rule SN_onE[OF SN], elim notE, intro exI conjI)
    from 0 show "(f \<circ> seq) 0 \<in> f ` X" by auto
    show "chain R (f \<circ> seq)"
    proof
      fix i
      show "((f \<circ> seq) i, (f \<circ> seq) (Suc i)) \<in> R" unfolding comp_def
      proof(rule stable)
        from 0 chain have "seq i \<in> (R ^^ i) `` X" by (induct i, auto)
        also have "... \<subseteq> R\<^sup>* `` X" by (rule Image_mono, auto simp: relpow_imp_rtrancl)
        finally show "seq i \<in> ..." using chain by auto
      qed (insert chain, auto)
    qed
  qed
qed

lemma (in itrans) i_transII:
  assumes sn: "SN_on (qrel Q R S) {t \<cdot> \<sigma>}"
  defines [simp]: "l \<equiv> \<lambda>t. i_trans (t \<cdot> \<sigma>)" and [simp]: "r \<equiv> \<lambda>t. t \<cdot> i_trans_subst \<sigma>"
  shows "(l t, r t) \<in> (rstep (ce_trs cn))\<^sup>* \<and>
    (is_Var t \<or> i_trans_check (R\<union>S) ur (t\<cdot>\<sigma>) \<or> (l t, r t) \<in> (rstep (ce_trs cn))\<^sup>+)"
  (is "?goal1 t \<and> (_ \<or> ?check (t\<cdot>\<sigma>) \<or> ?goal2 t)")
using sn
proof (induct t)
  case (Var x) show ?case by (auto simp: supteq_var_imp_eq)
next
  case (Fun f ts)
  then have SN: "SN_on (qrel Q R S) {Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ts)}" by auto
  then have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_on (qrel Q R S) {s \<cdot> \<sigma>}"
    by (simp add: SN_imp_SN_arg_gen[OF ctxt_closed_qrel])

  def t \<equiv> "Fun f ts"
  let ?rel = "rstep (ce_trs cn)"

  let ?us = "sorted_distinct_list (i_trans_sub ` (rstep S)\<^sup>* `` {t \<cdot> \<sigma>})"

  from Fun sns have nearlyDone: "(Fun f (map l ts), t \<cdot> i_trans_subst \<sigma>) \<in> ?rel\<^sup>*"
    unfolding subst_apply_term.simps t_def by(intro all_ctxt_closedD[of UNIV],auto)
  show ?case
  proof (cases "?check (t \<cdot> \<sigma>)")
    case False
    with Fun SN have id: "l t = (comb cn) ?us" by (auto simp add: o_def t_def)
    have "i_trans_sub (t \<cdot> \<sigma>) \<in> set ?us" using QT by(subst sorted_distinct_list, auto simp: Image_def)
    with id have "(l t, i_trans_sub (t \<cdot> \<sigma>)) \<in> (rstep (ce_trs cn))^+" by (auto intro: ce_trs_sound)
    moreover from SN False have "(i_trans_sub (t \<cdot> \<sigma>), Fun f (map l ts)) \<in> (rstep (ce_trs cn))^+"
      by(auto simp: t_def intro: ce_trs_sound)
    ultimately have "?goal2 t" using nearlyDone by auto
    then show ?thesis by (auto simp: t_def)
  next
    case check: True
      with Fun nearlyDone have *: "?goal1 t" by (simp add: o_def t_def)
      with check show ?thesis by (auto simp: t_def)
  qed
qed

lemma (in itrans) i_transIV_R:
  assumes sn: "SN_on (qrel Q R S) {s}"
      and st: "(s,t) \<in> qrstep False Q R"
      and ur: "\<not> i_trans_check (R\<union>S) ur s"
      and s: "is_Fun s"
  shows "(i_trans s, i_trans t) \<in> (rstep (ce_trs cn))^+" (is "_ \<in> ?rel")
proof -
  from s obtain f ss where s: "s = Fun f ss" by auto
  let ?us = "sorted_distinct_list (i_trans_sub ` (rstep S)\<^sup>* `` {s})"
  from ur sn s
  have "i_trans s = comb cn ?us" by auto
  also
    with QT have "i_trans_sub s \<in> set ?us" by (subst sorted_distinct_list, auto simp:Image_def)
    note ce_trs_sound[OF this]
  also from ur sn s
    have *: "i_trans_sub s = comb cn (Fun f (map i_trans ss) # map i_trans (qr_successors s))" by auto
    from sn have "SN_on (qrstep False Q R) {s}" by(rule SN_on_mono, auto simp: qrel_def)
    from st qr_successors[OF this]
    have "i_trans t \<in> set (map i_trans (qr_successors s))" by auto
    then have "(i_trans_sub s, i_trans t) \<in> ?rel" by(unfold *, auto intro: ce_trs_sound)
  ultimately show ?thesis by auto
qed



lemma (in itrans) i_transIV_S:
  assumes SNs: "SN_on (qrel Q R S) {s}"
      and st: "(s,t) \<in> rstep S"
      and ur: "\<not>i_trans_check (R\<union>S) ur s"
      and s: "is_Fun s"
  shows "(i_trans s, i_trans t) \<in> (rstep (ce_trs cn))\<^sup>*" (is "_ \<in> ?rel")
proof-
  from SN_on_relto_Image_S[OF SNs[unfolded qrel_def]] st have SNt: "SN_on (qrel Q R S) {t}"
    unfolding qrel_def by auto
  from s obtain f ss where "s = Fun f ss" by auto
  with ur SNs
  have 1: "i_trans s = comb cn (sorted_distinct_list (i_trans_sub ` (rstep S)\<^sup>* `` {s}))" (is "_ = ?i s") by auto
  show ?thesis
  proof(cases "i_trans_check (R\<union>S) ur t")
    case True
      from st QT have "(?i s, i_trans_sub t) \<in> (rstep (ce_trs cn))\<^sup>+"
        by(intro ce_trs_sound, subst sorted_distinct_list, auto simp: Image_def)
      moreover with True SNt have "(i_trans_sub t, i_trans t) \<in> ?rel" by (cases t, auto)
      ultimately show ?thesis using 1 by auto
  next
   case False
    with SNt have 2: "i_trans t = ?i t" by (cases t, auto)
    from QT st
    have "(i_trans s, i_trans t) \<in> ?rel"
      by(unfold 1 2, intro ce_trs_subset, (subst sorted_distinct_list, auto simp: Image_def)+)
    then show ?thesis by auto
  qed
qed


context ce_redpair
begin

definition c :: "'a" where "c \<equiv> undefined"
definition n :: nat where "n \<equiv> SOME n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c.  ce_trs (c,m) \<subseteq> NS)"

lemma NS_ce_compat_n: "m \<ge> n \<Longrightarrow> ce_trs (c,m) \<subseteq> NS"
proof -
  assume m: "n \<le> m"
  from mp[OF spec[OF someI_ex[OF NS_ce_compat]] m[unfolded n_def]] show "ce_trs (c,m) \<subseteq> NS" by blast
qed

lemma ce_orient: "m \<ge> n \<Longrightarrow> (rstep (ce_trs (c,m)))^* \<subseteq> NS^*"
  by (rule rtrancl_mono, rule rstep_subset, rule wm_NS, rule subst_NS, rule NS_ce_compat_n, simp)

abbreviation mono where "mono m \<equiv> ctxt.closed S \<and> ce_trs (c,m) \<subseteq> S"
abbreviation monoNS where "monoNS \<equiv> (NS \<union> S)^*"
abbreviation monoS where "monoS \<equiv> monoNS O S O monoNS"

lemma mono_ce: assumes m: "mono m"   
  shows "rstep (ce_trs (c,m)) \<subseteq> S"
  by (rule rstep_subset, insert m, auto simp: subst_S)

definition mode_cond where "mode_cond mmode m \<equiv> mmode \<longrightarrow> mono m"
definition mode_left where "mode_left mmode \<equiv> if mmode then NS \<union> S else NS"
definition mode_NS where "mode_NS mmode \<equiv> if mmode then monoNS else NS^*"

lemma mode_cond_left: "mode_cond mmode m \<Longrightarrow> ctxt.closed (mode_left mmode) \<and> subst.closed (mode_left mmode)"
  by (cases mmode,
    auto simp: mode_cond_def mode_left_def subst_S subst_NS wm_NS intro: subst.closed_Un mono_NSS)

lemma mode_left_NS: "mode_left mode \<subseteq> mode_NS mode"
  unfolding mode_left_def mode_NS_def
  by (cases mode, auto)

lemma mode_NS_mode[simp]: "NS^* O mode_NS mode = mode_NS mode"
proof (cases mode)
  case False thus ?thesis unfolding mode_left_def mode_NS_def by auto
next
  case True 
  hence id: "mode = True" by simp
  show ?thesis unfolding mode_left_def mode_NS_def id
    by (simp, regexp)
qed

lemma mode_mode_NS[simp]: "mode_NS mode O NS^* = mode_NS mode"
proof (cases mode)
  case False thus ?thesis unfolding mode_left_def mode_NS_def by auto
next
  case True 
  hence id: "mode = True" by simp
  show ?thesis unfolding mode_left_def mode_NS_def id
    by (simp, regexp)
qed
end

context ce_af_redpair begin

context
  fixes Q :: "('f,'v) term set" and R E :: "('f,'v) trs"
  assumes fin_R: "finite (applicable_rules Q R)"
  and wf_R: "wwf_qtrs Q R"
  and QT_E: "quasi_terminating (rstep E)"
  and no_left_var: "\<forall>(l,r) \<in> E. is_Fun l" (* not essential *)
begin

abbreviation "RE \<equiv> R \<union> E"

lemma itrans: "itrans Q R E" using QT_E fin_R wf_R unfolding itrans_def by auto

lemma i_transI:
  assumes ur: "ur_closed_term_af RE ur \<pi> t"
    and sn: "SN_on (qrel Q R E) {t \<cdot> \<sigma>}"
  shows "(t \<cdot> (itrans.i_trans_subst Q R E ur com \<sigma>), itrans.i_trans Q R E ur com (t \<cdot> \<sigma>)) \<in> NS^*"
proof -
  interpret it: itrans Q R E ur com by (rule itrans)
  from assms show ?thesis
  proof (induct t)
    case (Fun f ts)
    hence sClosed: "\<And> i. i < length ts \<and> i \<in> \<pi> (f, (length ts)) \<Longrightarrow> ur_closed_term_af RE ur \<pi> (ts ! i)" by auto
    from \<open>SN_on (qrel Q R E) {Fun f ts \<cdot> \<sigma>}\<close> have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_on (qrel Q R E) {s \<cdot> \<sigma>}"
      by (simp add: SN_imp_SN_arg_gen[OF ctxt_closed_qrel])
    have "it.i_trans (Fun f ts \<cdot> \<sigma>) = it.i_trans (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" (is "?l = it.i_trans ?r") by auto
    also have "\<dots> = Fun f (map (it.i_trans \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ts)"
    proof -
      from Fun(2)
      have inst: "\<forall> (l,r) \<in> RE. connectable RE (Fun f ts) l \<longrightarrow> (l,r) \<in> ur" by auto
      have "i_trans_check RE ur (Fun f ts \<cdot> \<sigma>)" (is ?check)
      proof
        fix l r assume lr: "(l,r) \<in> RE" and con: "connectable RE (Fun f ts \<cdot> \<sigma>) l"
        from connectable_stable[OF con] have "connectable RE (Fun f ts) l" by auto
        with inst lr show "(l,r) \<in> ur" by auto
      qed
      then have "i_trans_check RE ur ?r" by auto
      with Fun(3) show ?thesis by auto
    qed
    finally have eq1: "?l = Fun f (map (\<lambda> t. it.i_trans (t \<cdot> \<sigma>)) ts)" (is "_ = ?r") by simp
    let ?tis = "\<lambda> t. t \<cdot> it.i_trans_subst \<sigma>"
    let ?its = "\<lambda> t. it.i_trans (t \<cdot> \<sigma>)"
    have "(Fun f (map ?tis ts), Fun f (map ?its ts)) \<in> NS^*" 
    proof -
      {
        fix i
        assume ilen: "i < length ts" and ipi: "i \<in> \<pi> (f, (length ts))"
        with sns have "ts ! i \<in> set ts" and "SN_on (qrel Q R E) {ts ! i \<cdot> \<sigma>}" by auto
        with Fun sClosed ilen ipi have "(?tis (ts ! i), ?its (ts ! i)) \<in> NS^*" by auto
      } note ind = this
      from wm_NS have cc: "ctxt.closed (NS^*)" by blast
      show ?thesis 
      proof (rule af_steps_imp_orient, rule trans_rtrancl, rule refl_rtrancl, rule cc, force)
        show "\<forall>i<length (map ?tis ts). i \<in> \<pi> (f, (length ts)) \<longrightarrow> (map ?tis ts ! i, map ?its ts ! i) \<in> NS^*"
          using ind by auto
        show "\<forall>bef s t aft. length (map ?tis ts) = Suc (length bef + length aft) \<longrightarrow> length bef \<in> \<pi> (f, (length ts)) \<or> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> NS\<^sup>*"
          using af_compat unfolding af_compatible_def by force
      qed
    qed
    also have "Fun f (map ?tis ts) = Fun f ts \<cdot> it.i_trans_subst \<sigma>" by auto
    finally show ?case using eq1 by auto
  qed simp
qed

lemma SN_ac_preservation: assumes st: "(s,t) \<in> qrstep False Q R \<union> rstep E"
  and SN: "SN_on (qrel Q R E) {s}"
  shows "SN_on (qrel Q R E) {t}"
  by (rule step_preserves_SN_on_relto[OF _ SN[unfolded qrel_def], of t, folded qrel_def],
  insert st, auto)

lemma SN_ac_preservation_steps: assumes SN: "SN_on (qrel Q R E) {t}" 
  shows "(t,s) \<in> (qrstep False Q R \<union> rstep E)^* \<Longrightarrow> SN_on (qrel Q R E) {s}"
proof (induct rule: rtrancl_induct)
  case (step s r)
  from SN_ac_preservation[OF step(2-3)]
  show ?case by auto
qed (rule SN)

lemma i_transIII: assumes 
  mode: "mode_cond mode m" and
  m: "m \<ge> n" and
  sn: "SN_on (qrel Q R E) {Fun f ts}" and
  choice: "lr \<in> R \<and> Q' = Q \<or> lr \<in> E \<and> Q' = {}" and
  step: "(Fun f ts,s) \<in> rqrstep False Q' {lr}" and
  ur: "i_trans_check RE ur (Fun f ts)" and
  ur_closed: "ur_closed_af RE ur \<pi>" and
  orient: "ur \<subseteq> mode_left mode"
  shows "((itrans.i_trans Q R E ur (c,m) (Fun f ts), itrans.i_trans Q R E ur (c,m) s) \<in> mode_NS mode) \<and> 
    (mode \<and> lr \<in> S \<union> (R \<union> E - ur) \<longrightarrow> (itrans.i_trans Q R E ur (c,m) (Fun f ts), itrans.i_trans Q R E ur (c,m) s) \<in> monoS) \<and> lr \<in> ur"
proof -
  from step choice have rule: "lr \<in> RE" by auto
  let ?cn = "(c,m)"
  let ?QR = "qrel Q R E"
  interpret it: itrans Q R E ur ?cn by (rule itrans)
  let ?I = "it.i_trans"
  obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
  note rule = rule[unfolded lr]
  from step[unfolded lr rqrstep_def] obtain \<sigma> where left: "l \<cdot> \<sigma> = Fun f ts" and right: "r \<cdot> \<sigma> = s" by auto
  let ?Is = "it.i_trans_subst \<sigma>"
  show ?thesis
  proof(cases "is_Var l \<and> lr \<in> R")
    case True
    with lr step choice obtain x where "(Var x, r) \<in> R" by (cases l, auto)
    hence "\<not> (SN_on (qrstep False Q R) {Fun f ts})" 
      by (rule left_Var_imp_not_SN_qrstep, auto)
    with SN_on_mono[OF _ qrstep_subset_qrel, of Q R E "{Fun f ts}"]
    have False using sn by auto
    then show ?thesis by auto
  next
    case False
    from ur have nmatch: "\<forall> (l,r) \<in> RE - ur. \<not> connectable RE (Fun f ts) l" by (auto simp: i_trans_check_def)
    have "connectable RE (Fun f ts) l"
    proof(cases "l")
      case (Var x) then show ?thesis by(intro connectableI[of _ Var _ "\<lambda>x. Fun f ts"],auto)
    next
      case l: (Fun g ls)
      with left have fg: "f = g" and tsss: "ts = map (\<lambda> s. s \<cdot> \<sigma>) ls" (is "_ = ?ts") and nvar: "\<not> (is_Var l)" by auto
      from nmatch[unfolded tsss] have "\<forall> (l,r) \<in> RE - ur. \<not> connectable RE (Fun f ls \<cdot> \<sigma>) l" by auto
      have "connectable RE (Fun f ls \<cdot> \<sigma>) l" by (intro connectableI[of _ Var _ \<sigma>], auto simp: l fg)
      with tsss show ?thesis by auto
    qed
    with rule nmatch have urule: "(l, r) \<in> ur" by auto
    with left and right have urStep: "(Fun f ts, s) \<in> rstep ur" by (auto) (metis rstep.rule rstep.subst)
    from urule ur_closed have fr: "ur_closed_term_af RE ur \<pi> r" by auto
    from left sn have snl: "SN_on ?QR {l \<cdot> \<sigma>}" by auto
    with conjunct2[OF it.i_transII[OF snl]] False left no_left_var rule lr
    have oneB: "i_trans_check RE ur (l\<cdot>\<sigma>) \<or> (?I (Fun f ts), l \<cdot> ?Is) \<in> (rstep (ce_trs ?cn))^+" unfolding lr by auto
    from it.i_transII[OF snl] ce_orient[OF m] have "(?I (l \<cdot> \<sigma>), l \<cdot> ?Is) \<in> NS^*" by blast
    with left have one: "(?I (Fun f ts), l \<cdot> ?Is) \<in> NS^*" by auto
    from choice step have "(Fun f ts, s) \<in> rqrstep False Q R \<union> rqrstep False {} E" unfolding rqrstep_def
      by (cases, blast+)
    hence "(Fun f ts, s) \<in> qrstep False Q R \<union> qrstep False {} E" unfolding qrstep_iff_rqrstep_or_nrqrstep  by auto
    hence "(Fun f ts, s) \<in> qrstep False Q R \<union> rstep E" by auto
    from SN_ac_preservation[OF this sn] have sns: "SN_on ?QR {r \<cdot> \<sigma>}" unfolding right .
    have "(r \<cdot> ?Is , ?I (r \<cdot> \<sigma>)) \<in> NS^*"  by (rule i_transI[OF fr sns])
    with right have stepRight: "(r \<cdot> ?Is, ?I s) \<in> NS^*" by auto
    from urule have urstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep ur" by auto
    with rstep_subset[OF _ _ orient] mode_cond_left[OF mode] have lstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> mode_left mode" by auto
    with mode_left_NS have rstep: "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> mode_NS mode" by auto
    with stepRight have two: "(l \<cdot> ?Is, ?I s) \<in> mode_NS mode O NS^*" by blast
    from one two have "(?I (Fun f ts), ?I s) \<in> NS^* O mode_NS mode O NS^*" by blast
    hence part1: "(?I (Fun f ts), ?I s) \<in> mode_NS mode" by simp
    show ?thesis
    proof (intro conjI impI, rule part1, unfold lr, erule conjE)
      show "(l,r) \<in> ur" by (rule urule)
    next
      assume mono: mode and S: "(l,r) \<in> S \<union> (R \<union> E - ur)"
      from S have "(l,r) \<in> S \<or> (l,r) \<notin> ur" by auto
      hence "(?I (Fun f ts), r \<cdot> ?Is) \<in> monoS"
      proof
        assume "(l,r) \<in> S"
        with subst_S have "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> S" unfolding subst.closed_def by auto
        moreover have "(?I (Fun f ts), l \<cdot> ?Is) \<in> monoNS" using one rtrancl_mono by blast
        ultimately show ?thesis by auto
      next
        assume "(l, r) \<notin> ur" with urule show ?thesis by auto
      qed
      with stepRight have "(?I (Fun f ts), ?I s) \<in> monoS O NS^*" by blast
      moreover have "monoS O NS^* \<subseteq> monoS" by regexp
      ultimately show "(?I (Fun f ts), ?I s) \<in> monoS" by blast
    qed
  qed
qed

lemma mode_NS: "NS^* \<subseteq> mode_NS mode"
  unfolding mode_NS_def
  by (cases mode, insert rtrancl_mono[of NS], auto)

lemma mode_NS_ctxt: "mode_cond mode m \<Longrightarrow> ctxt.closed (mode_NS mode)"
proof (cases mode)
  case False thus ?thesis unfolding mode_NS_def using wm_NS by auto 
next
  case True
  assume "mode_cond mode m"
  hence "ctxt.closed S" unfolding mode_cond_def using True by auto
  from mono_NSS[OF this] show ?thesis unfolding mode_NS_def using True by auto 
qed

lemma mode_S_ctxt: "mode \<Longrightarrow> mode_cond mode m \<Longrightarrow> ctxt.closed monoS"
proof -
  assume m: mode and c: "mode_cond mode m"
  from mode_NS_ctxt[OF c] have NS: "ctxt.closed monoNS" unfolding mode_NS_def
    using m by auto
  from c m have S: "ctxt.closed S" unfolding mode_cond_def by auto
  from S NS show ?thesis by blast
qed

lemma mode_ce: "mode \<Longrightarrow> mode_cond mode m \<Longrightarrow> (rstep (ce_trs (c,m)))^+ \<subseteq> monoS"
proof -
  assume mode and mode: "mode_cond mode m"
  with mode have "mono m" unfolding mode_cond_def by auto
  from trancl_mono[OF _ mono_ce[OF this]] have "(rstep (ce_trs (c,m)))^+ \<subseteq> S^+" by auto
  moreover have "S^+ \<subseteq> monoS" by regexp
  ultimately show ?thesis by auto
qed

lemma i_transV:
  assumes mode: "mode_cond mode m"
  and m: "m \<ge> n"
  and choice: "lr \<in> R \<and> Q' = Q \<or> lr \<in> E \<and> Q' = {}"
  and sn: "SN_on (qrel Q R E) {t}"
  and step: "(t,s) \<in> qrstep False Q' {lr}"
  and ur_closed: "ur_closed_af RE ur \<pi>"
  and orient: "ur \<subseteq> mode_left mode"
  shows "(itrans.i_trans Q R E ur (c,m) t, itrans.i_trans Q R E ur (c,m) s) \<in> mode_NS mode \<and>
  (mode \<and> lr \<in> R \<and> Q' = Q \<and> (lr \<notin> ur \<or> lr \<in> S)
  \<longrightarrow> (itrans.i_trans Q R E ur (c,m) t, itrans.i_trans Q R E ur (c,m) s) \<in> monoS)"
  (is "_ \<and> (?cond \<longrightarrow> _)")
proof -
  let ?L = "mode_left mode"
  let ?R = "mode_NS mode"
  let ?cn = "(c,m)"
  let ?QR = "qrel Q R E"
  interpret it: itrans Q R E ur ?cn by (rule itrans)
  let ?I = "it.i_trans"
  let ?goal = "\<lambda> t s.  (?I t, ?I s) \<in> ?R \<and> (?cond \<longrightarrow> (?I t, ?I s) \<in> monoS)"
  from step choice have rule: "lr \<in> RE" by auto
  from step sn show "?goal t s"
  proof induct 
    case (IH C \<sigma> l r)
    with choice rule have lr: "lr = (l,r)" and IH2: "(l,r) \<in> RE" 
      and choice: "(l,r) \<in> R \<and> Q' = Q \<or> (l,r) \<in> E \<and> Q' = {}" by auto
    have sub: "\<And> x. x \<in> qrstep False Q' {(l,r)} \<Longrightarrow> x \<in> qrstep False Q R \<union> rstep E" using choice 
      by (cases, auto)
    from IH(4)
    show ?case
    proof (induct C)
      case Hole
      let ?ls = "\<box>\<langle>l \<cdot> \<sigma>\<rangle>" 
      let ?rs = "\<box>\<langle>r \<cdot> \<sigma>\<rangle>" 
      {
        assume "is_Var ?ls"
        then obtain x where "?ls = Var x" by (cases ?ls, auto)
        then obtain x where l: "l = Var x" by (cases l, auto)
        with choice no_left_var lr have "(l,r) \<in> R" by auto
        from left_Var_imp_not_SN_qrstep[OF this[unfolded l], of False] IH(4)
          SN_on_mono[OF _ qrstep_subset_qrel, of Q R E]
        have False by auto
      }
      hence "\<not> (is_Var ?ls)" ..
      from this obtain f ls where ls: "?ls = Fun f ls" by (cases ?ls, auto)
      from IH(3) Hole ls have snls: "SN_on ?QR {Fun f ls}" by auto
      from rqrstepI[OF IH(1-2) _ _ IH(3)] ls have rstep: "(Fun f ls, ?rs) \<in> rqrstep False Q' {(l,r)}"
        unfolding lr by auto
      hence step: "(Fun f ls, ?rs) \<in> qrstep False Q' {(l,r)}" unfolding qrstep_iff_rqrstep_or_nrqrstep by blast
      show "?goal ?ls ?rs"
      proof (cases "i_trans_check RE ur (Fun f ls)")
        case True
        from i_transIII[OF mode m snls choice rstep True ur_closed orient]
        show ?thesis using IH2 unfolding ls lr by blast
      next
        case False note check = this
        from choice
        show ?thesis
        proof (cases "(l,r) \<in> R \<and> Q' = Q")
          case True
          with qrstep_mono[of "{(l,r)}" R Q _ False, OF _ subset_refl] step
          have "(Fun f ls, ?rs) \<in> qrstep False Q R" by auto
          with snls check have steps: "(?I ?ls, ?I ?rs) \<in> (rstep (ce_trs ?cn))^+" unfolding ls
          by(intro it.i_transIV_R, auto)
          from steps ce_orient[OF m] have "(?I ?ls, ?I ?rs) \<in> NS^*" by auto
          hence piece0:"(?I ?ls, ?I ?rs) \<in> mode_NS mode" using mode_NS by blast
          show ?thesis 
            by (intro conjI[OF piece0] impI, elim conjE, insert steps mode_ce[OF _ mode], auto)
        next
          case False
          with choice have "(l,r) \<in> E \<and> Q' = {}" by auto
          with step have "(Fun f ls, ?rs) \<in> rstep E" by auto
          from it.i_transIV_S[OF snls this check] ce_orient m
          have "(?I ?ls, ?I ?rs) \<in> NS^*" unfolding ls by blast
          hence piece0:"(?I ?ls, ?I ?rs) \<in> mode_NS mode" using mode_NS by blast
          with False IH show ?thesis using ls by auto
        qed
      qed
    next
      case (More f bef D aft)
      obtain t s where t: "t = l \<cdot> \<sigma>" and s: "s = r \<cdot> \<sigma>" by auto
      from More
      have snbta: "SN_on ?QR {Fun f (bef @ D\<langle>t\<rangle> # aft)}" (is "SN_on _ {Fun f ?bta}") 
        unfolding t by auto
      let ?Dt = "D\<langle>t\<rangle>"
      let ?Ds = "D\<langle>s\<rangle>"
      let ?bsa = "bef @ ?Ds # aft"
      let ?C = "More f bef D aft"
      have step: "(Fun f ?bta, Fun f ?bsa) \<in> qrstep False Q' {(l,r)}" 
        unfolding s t
        by (rule qrstepI[OF IH(1) _ _ _ IH(3), of _ _ ?C], auto)
      show ?case 
      proof (cases "i_trans_check RE ur (Fun f ?bta)")
        case False note check = this
        let ?l = "Fun f ?bta"
        let ?r = "Fun f ?bsa"
        show ?thesis
        proof (cases "(l,r) \<in> R \<and> Q' = Q")
          case True
          with qrstep_mono[of "{(l,r)}" R Q _ False, OF _ subset_refl] step
          have "(?l, ?r) \<in> qrstep False Q R" by auto
          from snbta this check
          have steps: "(?I ?l, ?I ?r) \<in> (rstep (ce_trs ?cn))^+" by (intro it.i_transIV_R, auto)
          from steps ce_orient[OF m] have "(?I ?l, ?I ?r) \<in> NS^*" by auto
          hence piece0: "(?I ?l, ?I ?r) \<in> mode_NS mode" using mode_NS by blast
          show ?thesis unfolding s[symmetric] t[symmetric]
            by (rule conjI, insert piece0, simp, intro impI, elim conjE, insert steps mode_ce[OF _ mode], auto)
        next
          case False
          with choice have "(l,r) \<in> E \<and> Q' = {}" by auto
          with step have "(?l,?r) \<in> rstep E" by auto
          from it.i_transIV_S[OF snbta this check] ce_orient[OF m] have "(?I ?l, ?I ?r) \<in> NS^*" by auto
          hence piece0:"(?I ?l, ?I ?r) \<in> mode_NS mode" using mode_NS by blast
          with False IH show ?thesis unfolding s[symmetric] t[symmetric] by auto
        qed
      next
        case True
        have "?Dt \<in> set ?bta" by auto 
        with snbta have "SN_on ?QR {?Dt}" by (rule SN_imp_SN_arg_gen[OF ctxt_closed_qrel])
        with More have "?goal ?Dt ?Ds" unfolding s t by auto
        hence always: "(?I ?Dt, ?I ?Ds) \<in> ?R" and choice: "?cond \<longrightarrow> (?I ?Dt, ?I ?Ds) \<in> monoS" by auto
        from ctxt_closed_one[OF mode_NS_ctxt[OF mode] always]
        have piece0: "(Fun f (map ?I bef @ ?I ?Dt # map ?I aft),
          Fun f (map ?I bef @ ?I ?Ds # map ?I aft)) \<in> ?R" (is "(?Il, ?Ir) \<in> _") .
        from choice have choice: "?cond \<longrightarrow> (?Il, ?Ir) \<in> monoS" using ctxt_closed_one[OF mode_S_ctxt[OF _ mode]] by blast
        from True snbta have left: "?I (Fun f ?bta) = ?Il" by simp
        have snbsa: "SN_on ?QR {Fun f ?bsa}" by (rule SN_ac_preservation[OF sub[OF step] snbta])
        from IH2 ctxt.closure.intros have "(?Dt,?Ds) \<in> rstep RE"
          unfolding rstep_ctxt_closure_subst_closure s t by auto
        have "i_trans_check RE ur (Fun f ?bsa)"
        proof(rule i_trans_check_nrrstep[OF ur_closed True])
          from s t lr rule have "(t,s)\<in>rstep RE" by auto
          from rstep_ctxt_imp_nrrstep[OF this, of ?C]
          show "(Fun f ?bta, Fun f ?bsa) \<in> nrrstep RE" by auto
        qed auto
        with snbsa have "?I (Fun f ?bsa) = ?Ir" by auto
        with left piece0 choice show ?thesis unfolding s t by simp
      qed
    qed
  qed
qed

lemma i_transV_R:
  assumes mode: "mode_cond mode m"
  and m: "m \<ge> n"
  and lr: "lr \<in> R"
  and sn: "SN_on (qrel Q R E) {t}"
  and step: "(t,s) \<in> qrstep False Q {lr}"
  and ur_closed: "ur_closed_af RE ur \<pi>"
  and orient: "ur \<subseteq> mode_left mode"
  shows "(itrans.i_trans Q R E ur (c,m) t, itrans.i_trans Q R E ur (c,m) s) \<in> mode_NS mode"
    and "mode \<Longrightarrow> lr \<notin> ur \<or> lr \<in> S \<Longrightarrow> (itrans.i_trans Q R E ur (c,m) t, itrans.i_trans Q R E ur (c,m) s) \<in> monoS"
  using i_transV[OF assms(1,2) _ assms(4-7)] lr by auto

lemma i_transV_E:
  assumes mode: "mode_cond mode m"
  and m: "m \<ge> n"
  and choice: "lr \<in> E"
  and sn: "SN_on (qrel Q R E) {t}"
  and step: "(t,s) \<in> rstep {lr}"
  and ur_closed: "ur_closed_af RE ur \<pi>"
  and orient: "ur \<subseteq> mode_left mode"
  shows "(itrans.i_trans Q R E ur (c,m) t, itrans.i_trans Q R E ur (c,m) s) \<in> mode_NS mode"
  using i_transV[OF assms(1,2) _ assms(4) _ assms(6-7), of "lr" "{}"] assms(3) assms(5) by auto

lemma i_transVI: assumes
  mode: "mode_cond mode m" and
  m: "m \<ge> n" and
  sn: "SN_on (qrel Q R E) {t}" and
  step: "(t,s) \<in> (qrstep False Q R \<union> rstep E)^*" and
  ur_closed: "ur_closed_af RE ur \<pi>" and
  orient: "ur \<subseteq> mode_left mode"
  shows "(itrans.i_trans Q R E ur (c,m) t, itrans.i_trans Q R E ur (c,m) s) \<in> mode_NS mode"
 (is "?goal t s")
using step 
proof (induct rule: rtrancl_induct)
  case base
  show ?case by (cases mode, unfold mode_NS_def, auto)
next
  case (step s u)
  let ?cn = "(c,m)"
  let ?QR = "qrel Q R E"
  interpret it: itrans Q R E ur ?cn by (rule itrans)
  let ?I = "it.i_trans"
  from step(1-2) have "(t,s) \<in> (qrstep False Q R \<union> rstep E)^*" by auto
  from SN_ac_preservation_steps[OF sn this] have sn': "SN_on (qrel Q R E) {s}" .
  from step(2) have isteps2: "(?I s, ?I u) \<in> mode_NS mode"
  proof 
    assume "(s, u) \<in> qrstep False Q R"
    from this[unfolded qrstep_rule_conv[where R = R]] obtain lr where lr: "lr \<in> R" 
      and sustep: "(s,u) \<in> qrstep False Q {lr}" by auto
    from i_transV[OF mode m _ sn' sustep ur_closed orient] lr show ?thesis by auto
  next
    assume "(s,u) \<in> rstep E"
    from this obtain lr where lr: "lr \<in> E"
      and sustep: "(s,u) \<in> qrstep False {} {lr}" by blast
    from i_transV[OF mode m _ sn' sustep ur_closed orient] lr show ?thesis by auto
  qed
  with step(3) have "(?I t, ?I u) \<in> mode_NS mode O mode_NS mode" by auto
  thus ?case unfolding mode_NS_def by (cases mode, auto)
qed

lemma i_trans_strict_step:
  assumes mode: "mode_cond mode m"
  and modeT: "mode"
  and m: "m \<ge> n"
  and SNt: "SN_on (qrel Q R E) {t \<cdot> \<sigma>}"
  and step: "(t \<cdot> \<sigma> ,s \<cdot> \<delta>) \<in> (qrstep False Q R \<union> rstep E)^* O qrstep False Q R' O (qrstep False Q R \<union> rstep E)^*"
  and ur_closed: "ur_closed_af RE ur \<pi>"
  and ur_P_closed: "ur_P_closed_af RE ur \<pi> P"
  and st: "(s',t) \<in> P"
  and orient: "ur \<subseteq> mode_left mode"
  and R': "R' \<inter> ur \<subseteq> S"
  and RR': "R' \<subseteq> R"  
  shows "(t \<cdot> itrans.i_trans_subst Q R E ur (c,m) \<sigma>, s \<cdot> itrans.i_trans_subst Q R E ur (c,m) \<delta>) \<in> monoS"
proof -
  let ?cn = "(c,m)"
  let ?QR = "qrel Q R E"
  interpret it: itrans Q R E ur ?cn by (rule itrans)
  let ?I = "it.i_trans"
  let ?\<sigma> = "it.i_trans_subst \<sigma>"
  let ?\<delta> = "it.i_trans_subst \<delta>"
  from st ur_P_closed have urt: "ur_closed_term_af RE ur \<pi> t" by auto
  from i_transI[OF urt SNt] have ns0: "(t \<cdot> ?\<sigma>, ?I (t \<cdot> \<sigma>)) \<in> NS^*" . 
  note step
  also have "(qrstep False Q R \<union> rstep E)^* O qrstep False Q R' O (qrstep False Q R \<union> rstep E)^* \<subseteq> (qrstep False Q R \<union> rstep E)^* O qrstep False Q R O (qrstep False Q R \<union> rstep E)^*"
    by (rule relto_mono[OF qrstep_mono[OF RR']], auto)
  also have "\<dots> \<subseteq> (qrstep False Q R \<union> rstep E)^*" by regexp
  finally have "(t \<cdot> \<sigma>, s \<cdot> \<delta>) \<in> (qrstep False Q R \<union> rstep E)^*" .
  from SN_ac_preservation_steps[OF SNt this] have SNs: "SN_on (qrel Q R E) {s \<cdot> \<delta>}" .
  from it.i_transII[OF SNs] ce_orient[OF m] have ns2: "(?I (s \<cdot> \<delta>), s \<cdot> ?\<delta>) \<in> NS^*" by auto
  from step obtain u v
  where tu: "(t\<cdot>\<sigma>,u) \<in> (qrstep False Q R \<union> rstep E)\<^sup>*" 
    and uv: "(u,v) \<in> qrstep False Q R'"
    and vs: "(v,s\<cdot>\<delta>) \<in> (qrstep False Q R \<union> rstep E)\<^sup>*" by auto
  from SN_ac_preservation_steps[OF SNt] tu have SNu: "SN_on (qrel Q R E) {u}" by auto
  then have SNv: "SN_on (qrel Q R E) {v}" by(rule SN_ac_preservation_steps, insert uv RR', blast)
  from i_transVI[OF mode m SNt tu ur_closed orient] RR' modeT R'
  have 1: "(?I (t\<cdot>\<sigma>), ?I u) \<in> monoNS" by (simp add: mode_NS_def)
  moreover
    from uv obtain lr
    where lr: "lr \<in> R'" and uv': "(u,v) \<in> qrstep False Q {lr}" using qrstep_rule_conv by blast
    with RR' have lrR: "lr \<in> R" by auto
    from i_transV_R(2)[OF mode m lrR SNu uv' ur_closed orient modeT] lr R'
    have 2: "(?I u, ?I v) \<in> monoNS O S O monoNS" by auto
  moreover
    from i_transVI[OF mode m SNv vs ur_closed orient] RR' modeT R'
    have 3: "(?I v, ?I (s\<cdot>\<delta>)) \<in> monoNS" by (simp add: mode_NS_def)
  ultimately have "(?I (t\<cdot>\<sigma>), ?I (s\<cdot>\<delta>)) \<in> monoNS O monoNS O S O monoNS O monoNS" by blast
  also have "... \<subseteq> monoNS O S O monoNS" by regexp
  finally have "(?I (t\<cdot>\<sigma>), ?I (s\<cdot>\<delta>)) \<in> ..." by auto
  with ns0 ns2 have "(t \<cdot> ?\<sigma>, s \<cdot> ?\<delta>) \<in> NS^* O monoNS O S O monoNS O NS^*" by auto
  thus "(t \<cdot> ?\<sigma>, s \<cdot> ?\<delta>) \<in> monoS" by regexp
qed
end
end

no_notation Ground_Context.equiv_class ("[_]")

lemma (in ce_af_redpair) i_trans_sound_dp: 
  assumes fin_R: "finite R"
  and wf: "wwf_qtrs Q R" (* required from previous lemmas *)
  and mode: "mode_cond mode m"
  and m: "m \<ge> n"
  and P: "(s,t) \<in> P"
  and steps: "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep nfs Q R)^*"
  and SN: "SN_on (qrstep nfs Q R) {t \<cdot> \<sigma>}"
  and ur_closed: "ur_closed_af R ur \<pi>"
  and ur_P_closed: "ur_P_closed_af R ur \<pi> P"
  and orient: "ur \<subseteq> mode_left mode"
  shows  "(t \<cdot> itrans.i_trans_subst Q R {} ur (c,m) \<sigma>, 
    u \<cdot> itrans.i_trans_subst Q R {} ur (c,m) \<tau>) \<in> mode_NS mode"
proof -
  let ?cn = "(c,m)"
  have fin2: "finite (applicable_rules Q R)"
    by (rule finite_subset, rule applicable_rules_subset, rule fin_R)
  then interpret it: itrans Q R "{}" ur ?cn by (rule itrans, insert wf, auto)
  let ?s = "s \<cdot> \<sigma>"
  let ?t = "t \<cdot> \<sigma>"
  let ?u = "u \<cdot> \<tau>"
  let ?I = "it.i_trans"
  let ?Is = "it.i_trans_subst"
  let ?ssig = "s \<cdot> ?Is \<sigma>"
  let ?tsig = "t \<cdot> ?Is \<sigma>"
  let ?utau = "u \<cdot> ?Is \<tau>"
  let ?QR = "qrel Q R {}"
  have switch': "qrstep False Q R = ?QR" unfolding qrel_empty_is_qrstep ..
  have switch: "qrstep nfs Q R = qrstep False Q R"
    by (rule wwf_qtrs_imp_nfs_False_switch, insert wf)
  note steps = steps[unfolded switch]
  note SN = SN[unfolded switch]
  note SN' = SN[unfolded switch']
  from i_transVI[OF fin2 wf it.QT] mode m SN' _ ur_closed orient steps
  have one: "(?I ?t, ?I ?u) \<in> mode_NS mode" by simp
  from P ur_P_closed have "ur_closed_term_af R ur \<pi> t" by auto
  from i_transI[OF fin2 wf it.QT] this SN' have two: "(?tsig, ?I ?t) \<in> NS^*" by simp
  from steps SN have sns: "SN_on (qrstep False Q R) {?u}" by (rule steps_preserve_SN_on[of _ _ "qrstep False Q R"])
  note sns' = sns[unfolded switch']
  from it.i_transII[OF sns', THEN conjunct1] ce_orient[OF m]
  have three: "(?I ?u, ?utau ) \<in> NS^*" by auto    
  from one two three have "(?tsig,?utau) \<in> NS^* O mode_NS mode O NS^*" by blast
  thus ?thesis by simp
qed

lemma (in mono_ce_af_redtriple) mono_redpair_ur_sound:
  assumes fin: "finite (R \<union> Rw)"
  and wf: "wwf_qtrs Q (R \<union> Rw)"
  and mono: "ctxt.closed S"
  and rules: "ur \<subseteq> NS \<union> S"
  and nonStrict: "P \<union> Pw \<subseteq> NS \<union> S"
  and strictP: "Ps \<subseteq> S"
  and strictR: "(Rs \<inter> ur) \<subseteq> S"
  and ur_closed: "ur_closed_af (R \<union> Rw) ur \<pi>"
  and ur_P_closed: "ur_P_closed_af (R \<union> Rw) ur \<pi> (P \<union> Pw)"
  and m: m
  and minChain: "min_ichain (nfs,m,P,Pw,Q,R,Rw) s t \<sigma>"
  shows  "\<exists> j. min_ichain (nfs,m,P - Ps, Pw - Ps, Q, R - Rs, Rw - Rs) (shift s j) (shift t j) (shift \<sigma> j)"
proof (rule min_ichain_split[OF minChain], rule)
  {
    fix RR
    assume RR: "RR \<subseteq> R \<union> Rw"
    from RR wf have "nfs \<Longrightarrow> Q \<noteq> {} \<Longrightarrow> wwf_qtrs Q RR" unfolding wwf_qtrs_def by blast
    from wwf_qtrs_imp_nfs_False_switch[OF this] have "qrstep nfs Q RR = qrstep False Q RR" .
  } note switch_gen = this
  from switch_gen[OF subset_refl] have switch: "qrstep nfs Q (R \<union> Rw) = qrstep False Q (R \<union> Rw)" .
  assume chain: "min_ichain (nfs,m,Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, Rs \<inter> (R \<union> Rw), R \<union> Rw - Rs) s t \<sigma>"
  hence chain: "min_ichain (nfs,m,Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, Rs \<inter> (R \<union> Rw), R \<union> Rw - Rs) s t \<sigma>" by auto    
  have id: "Ps \<inter> (P \<union> Pw) \<union> (P \<union> Pw - Ps) = P \<union> Pw" by auto
  have idR: "Rs \<inter> (R \<union> Rw) \<union> (R \<union> Rw - Rs) = R \<union> Rw" by auto
  have idS: "S \<union> NS = NS \<union> S" by auto
  from S_ce_compat obtain n' where S_compat: "\<And> m. m \<ge> n' \<Longrightarrow> ce_trs (c,m) \<subseteq> S" by blast
  obtain k where k: "k = max n n'" by auto
  hence n: "n \<le> k"  by simp
  obtain mode where mode: "mode" by blast
  from S_compat k mode have cond: "mode_cond mode k" unfolding mode_cond_def using mono by auto
  have fin2: "finite (applicable_rules Q (R \<union> Rw))" by(rule finite_subset, rule applicable_rules_subset, insert fin)
  interpret it: itrans Q "R \<union> Rw" "{}" ur "(c, k)" by(rule itrans[OF fin2 wf], auto)
  obtain \<tau> where \<tau>: "\<tau> = (\<lambda> i. it.i_trans_subst (\<sigma> i))" by auto
  from mode rules have orient: "ur \<subseteq> mode_left mode" unfolding mode_left_def by auto
  note chain = chain[unfolded min_ichain.simps ichain.simps minimal_cond_def id idR]
  {
    fix i
    from chain \<open>m\<close> switch
    have P: "(s i, t i) \<in> P \<union> Pw" and R: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep False Q (R \<union> Rw))^*" and 
    SN: "SN_on (qrstep False Q (R \<union> Rw)) {t i \<cdot> \<sigma> i}" by auto
    from i_trans_sound_dp[OF fin wf cond n P R SN ur_closed ur_P_closed orient]
    have "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (NS \<union> S)^*" unfolding \<tau> mode_NS_def using mode by auto
    hence "(t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> (rstep (NS \<union> S))^*" 
      using rtrancl_mono[OF subset_rstep[of "NS \<union> S"]] by auto 
  } note steps = this
  have chain: "ichain (nfs,m,S,NS,{},S,NS) s t \<tau>"
    unfolding ichain.simps qrstep_rstep_conv idS
  proof (intro conjI allI)
    fix i
    from chain nonStrict show "(s i, t i) \<in> NS \<union> S" by auto 
  next
    let ?Q = "qrstep False Q (R \<union> Rw)"
    let ?QQ = "relto (qrstep False Q (Rs \<inter> (R \<union> Rw))) (?Q \<union> rstep {})"
    let ?R = "rstep (NS \<union> S)"
    let ?RR = "?R^* O rstep S O ?R^*"
    let ?QR = "qrel Q (R \<union> Rw) {}"
    have switch': "qrstep False Q (R \<union> Rw) = ?QR" unfolding qrel_empty_is_qrstep ..
    show "(INFM i. (s i, t i) \<in> S) \<or> 
      (INFM i. (t i \<cdot> \<tau> i, s (Suc i) \<cdot> \<tau> (Suc i)) \<in> ?RR)" unfolding INFM_disj_distrib[symmetric] 
      unfolding INFM_nat
    proof (intro allI)
      fix m
      from chain 
      have "(INFM i. (s i, t i) \<in> Ps \<inter> (P \<union> Pw)) \<or>
        (INFM i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?QQ)" by (auto simp: switch_gen)
      from this[unfolded INFM_disj_distrib[symmetric], unfolded INFM_nat]
      obtain l where l: "Suc m < l" and alt:"(s l, t l) \<in> Ps \<inter> (P \<union> Pw) \<or> 
        (t l \<cdot> \<sigma> l, s (Suc l) \<cdot> \<sigma> (Suc l)) \<in> ?QQ" by blast
      hence l': "m < l" by auto
      let ?I = "it.i_trans"
      from chain m switch have SN: "\<And> l. SN_on (qrel Q (R \<union> Rw) {}) {t l \<cdot> \<sigma> l}" by (simp add: qrel_def)
      from alt show "\<exists> l > m. (s l, t l) \<in> S \<or> (t l \<cdot> \<tau> l, s (Suc l) \<cdot> \<tau> (Suc l)) \<in> ?RR"
      proof
        assume Ssteps: "(t l \<cdot> \<sigma> l, s (Suc l) \<cdot> \<sigma> (Suc l)) \<in> ?QQ"
        let ?tt = "t l \<cdot> \<tau> l"
        let ?st = "s (Suc l) \<cdot> \<tau> (Suc l)"
        from chain have P: "(s l, t l) \<in> P \<union> Pw" by auto
        let ?Rs = "Rs \<inter> (R \<union> Rw)"
        have "(?tt,?st) \<in> monoS" unfolding \<tau>
          apply (rule i_trans_strict_step[OF fin2 wf it.QT _ cond mode n SN Ssteps _ _ P orient])
using ur_closed ur_P_closed P strictR by auto
        with rtrancl_mono[OF subset_rstep[of "NS \<union> S"]] subset_rstep[of S] have "(?tt, ?st) \<in> ?RR" by auto
        with l' show ?thesis by auto
      next
        assume P: "(s l, t l) \<in> Ps \<inter> (P \<union> Pw)"
        with strictP have "(s l, t l) \<in> S" by force
        with l' show ?thesis by auto
      qed
    qed
  qed (auto simp: steps)
  have idS: "S \<inter> (NS \<union> S) = S" by auto
  from ichain_mono[OF chain subset_refl _ subset_refl subset_refl, of "NS \<union> S - S" "NS \<union> S - S"] 
    mono_redpair_sound_ichain[OF mono subset_refl subset_refl, of nfs m "{}" s t \<tau>]
  show False unfolding idS by blast
qed



end
